# test builtin delattr
try:
    delattr
except:
    print("SKIP")
    raise SystemExit

class A: pass
a = A()
a.x = 1
print(a.x)

delattr(a, 'x')

try:
    a.x
except AttributeError:
    print('AttributeError')

try:
    delattr(a, 'x')
except AttributeError:
    print('AttributeError')



# test builtin pow() with integral values
# 3 arg version

try:
    print(pow(3, 4, 7))
except NotImplementedError:
    print("SKIP")
    raise SystemExit

print(pow(555557, 1000002, 1000003))

# Tests for 3 arg pow with large values

# This value happens to be prime
x = 0xd48a1e2a099b1395895527112937a391d02d4a208bce5d74b281cf35a57362502726f79a632f063a83c0eba66196712d963aa7279ab8a504110a668c0fc38a7983c51e6ee7a85cae87097686ccdc359ee4bbf2c583bce524e3f7836bded1c771a4efcb25c09460a862fc98e18f7303df46aaeb34da46b0c4d61d5cd78350f3edb60e6bc4befa712a849
y = 0x3accf60bb1a5365e4250d1588eb0fe6cd81ad495e9063f90880229f2a625e98c59387238670936afb2cafc5b79448e4414d6cd5e9901aa845aa122db58ddd7b9f2b17414600a18c47494ed1f3d49d005a5

print(hex(pow(2, 200, x))) # Should not overflow, just 1 << 200
print(hex(pow(2, x-1, x))) # Should be 1, since x is prime
print(hex(pow(y, x-1, x))) # Should be 1, since x is prime
print(hex(pow(y, y-1, x))) # Should be a 'big value'
print(hex(pow(y, y-1, y))) # Should be a 'big value'



a = (1, 2, 3)
print(a.index(1))
print(a.index(2))
print(a.index(3))
print(a.index(3, 2))
try:
    print(a.index(3, 2, 2))
except ValueError:
    print("Raised ValueError")
else:
    print("Did not raise ValueError")

a = a + a
b = (0, 0, a)
print(a.index(2))
print(b.index(a))
print(a.index(2, 2))

try:
    a.index(2, 2, 2)
except ValueError:
    print("Raised ValueError")
else:
    print("Did not raise ValueError")



# test string modulo formatting with int values

try:
    '' % ()
except TypeError:
    print("SKIP")
    raise SystemExit

# basic cases
print("%d" % 10)
print("%+d" % 10)
print("% d" % 10)
print("%d" % -10)
print("%d" % True)
print("%i" % -10)
print("%i" % True)
print("%u" % -10)
print("%u" % True)
print("%x" % 18)
print("%o" % 18)
print("%X" % 18)
print("%#x" % 18)
print("%#X" % 18)
print("%#6o" % 18)
print("%#6x" % 18)
print("%#06x" % 18)

# with *
print("%*d" % (5, 10))
print("%*.*d" % (2, 2, 20))
print("%*.*d" % (5, 8, 20))

# precision
for val in (-12, 12):
    print(">%8.4d<" % val)
    print(">% 8.4d<" % val)
    print(">%+8.4d<" % val)
    print(">%08.4d<" % val)
    print(">%-8.4d<" % val)
    print(">%-08.4d<" % val)
    print(">%-+08.4d<" % val)

# test + option with various amount of padding
for pad in ('', ' ', '0'):
    for n in (1, 2, 3):
        for val in (-1, 0, 1):
            print(('%+' + pad + str(n) + 'd') % val)



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

# This script renders a graph of the CircuitPython rom image.
# It takes the single elf file and uses objdump to get its contents.

import pygraphviz as pgv
import click
import sh

# Replace dashes with underscores
objdump = sh.arm_none_eabi_objdump


def parse_hex(h):
    return int("0x" + h, 0)


BAD_JUMPS = ["UNPREDICTABLE", "_etext"]

SPECIAL_NODE_COLORS = {"main": "pink", "exception_table": "green"}


@click.command()
@click.argument("elf_filename")
def do_all_the_things(elf_filename):
    symbol = None
    last_address = 0
    all_symbols = {}
    symbols_by_debug_address = {}
    symbols_by_memory_address = {}
    symbols_by_linkage_name = {}
    # Gather type info so we know how to treat the disassembly
    debug_dump = objdump("--dwarf=info", elf_filename)
    debug_dump_lines = debug_dump.stdout.decode("utf-8").split("\n")
    symbol_stack = []
    symbol = None
    ignore = False
    min_call_site_param = 0x20000000
    for line in debug_dump_lines:
        if not line:
            continue
        parts = line.split()
        if line[1] == "<":
            if parts[-1] == "0":
                symbol = symbol_stack.pop()
                continue
            debug_type = parts[-1].strip("()")
            ignore = False
            # skip info about function parameters
            if debug_type == "DW_TAG_formal_parameter":
                ignore = True
            depth = int(parts[0].split(">")[0].strip("<"))
            if len(symbol_stack) == (depth - 1) and depth > 0:
                symbol_stack.append(symbol)
            elif symbol and "name" in symbol:
                if symbol["debug_type"] == "DW_TAG_variable":
                    if "start_address" not in symbol:
                        pass
                    else:
                        symbols_by_memory_address[symbol["start_address"]] = symbol
                elif symbol["debug_type"] in [
                    "DW_TAG_member",
                    "DW_TAG_label",
                    "DW_TAG_typedef",
                    "DW_TAG_enumerator",
                    "DW_TAG_enumeration_type",
                    "DW_TAG_base_type",
                    "DW_TAG_structure_type",
                    "DW_TAG_compile_unit",
                    "DW_TAG_union_type",
                ]:
                    # skip symbols that don't end up in memory. the type info is available through the debug address map
                    pass
                else:
                    if symbol["name"] in all_symbols:
                        # print(depth, symbol["name"])
                        # print(symbol)
                        # print(all_symbols[symbol["name"]])
                        # print()
                        pass
                    all_symbols[symbol["name"]] = symbol
            elif (
                symbol
                and symbol["debug_type"] == "DW_TAG_GNU_call_site_parameter"
                and "call_site_value" in symbol
            ):
                parent = -1
                while symbol_stack[parent]["debug_type"] != "DW_TAG_subprogram":
                    parent -= 1
                parent = symbol_stack[parent]

                # Only point to ROM
                addr = symbol["call_site_value"]
                if 0x2000 <= addr < 0x20000000:
                    if "outgoing_pointers" not in parent:
                        parent["outgoing_pointers"] = set()
                    parent["outgoing_pointers"].add(addr)
                    if addr not in symbols_by_memory_address:
                        symbols_by_memory_address[addr] = symbol
                        min_call_site_param = min(addr, min_call_site_param)
                        symbol["name"] = "name{:x}".format(addr)
            address = parse_hex(parts[0].split("<")[-1].strip(">:"))
            symbol = {"debug_address": address, "debug_type": debug_type, "other": []}
            if debug_type == "DW_TAG_structure_type":
                symbol["struct"] = {}
            elif debug_type == "DW_TAG_array_type":
                symbol["subtype"] = None
                symbol["bound_count"] = 0
                symbol["maxlen"] = 0
            elif debug_type == "DW_TAG_subrange_type":
                symbol_stack[-1]["subtype"] = symbol
            symbols_by_debug_address[address] = symbol
        elif ignore:
            continue
        elif line[:4] == "    ":
            tag = parts[1].strip(":")
            if tag == "DW_AT_name":
                symbol["name"] = parts[-1]
            elif tag == "DW_AT_type":
                symbol["type"] = int(parts[-1].strip("<>"), 0)
                if symbol["debug_type"] == "DW_TAG_subrange_type":
                    if not symbol_stack[-1]["subtype"]:
                        symbol_stack[-1]["subtype"] = symbol
                    elif symbol_stack[-1]["subtype"]["type"] == symbol["type"]:
                        second_subtype = True
                    else:
                        raise RuntimeError()
            elif tag == "DW_AT_upper_bound":
                # Skip arrays with length defined by other variables
                if parts[-1][0] != "<":
                    upper_bound = int(parts[-1])
                    if symbol_stack[-1]["bound_count"] > 0:
                        symbol_stack[-1]["maxlen"] *= upper_bound + 1
                    else:
                        symbol_stack[-1]["maxlen"] = upper_bound + 1
                    symbol_stack[-1]["bound_count"] += 1
            elif tag == "DW_AT_byte_size":
                symbol["size"] = int(parts[-1])
            elif tag == "DW_AT_inline":
                symbol["inlined"] = True
            elif tag == "DW_AT_low_pc":
                addr = int(parts[-1], 0)
                symbols_by_memory_address[addr] = symbol
            elif tag == "DW_AT_location":
                if parts[-2] == "(DW_OP_addr:":
                    addr = parse_hex(parts[-1].strip(")"))
                    if addr > 0:
                        symbol["start_address"] = addr
            elif tag == "DW_AT_linkage_name":
                symbol["linkage_name"] = parts[-1]
                symbols_by_linkage_name[symbol["linkage_name"]] = symbol
            elif tag == "DW_AT_data_member_location":
                symbol_stack[-1]["struct"][int(parts[-1])] = symbol
            elif tag == "DW_AT_GNU_call_site_value":
                if parts[-2] == "(DW_OP_addr:":
                    symbol["call_site_value"] = parse_hex(parts[-1].strip(")"))
            else:
                symbol["other"].append(line)
                # print(parts)
                pass
        else:
            # print(line)
            pass

    MEMORY_NONE = 0
    MEMORY_POINTER = 1
    MEMORY_PY_OBJECT = 2

    def get_size(t):
        if "size" in t:
            return t["size"]
        return get_size(symbols_by_debug_address[t["type"]])

    def get_pointer_map(t, depth=0):
        if t["debug_type"] == "DW_TAG_pointer_type":
            return {0: MEMORY_POINTER}
        elif t["debug_type"] in [
            "DW_TAG_const_type",
            "DW_TAG_typedef",
            "DW_TAG_member",
            "DW_TAG_subrange_type",
            "DW_TAG_volatile_type",
        ]:
            if "name" in t and t["name"] == "mp_rom_obj_t":
                return {0: MEMORY_PY_OBJECT}
            return get_pointer_map(symbols_by_debug_address[t["type"]], depth + 1)
        elif t["debug_type"] in ["DW_TAG_base_type", "DW_TAG_enumeration_type"]:
            return {}
        elif t["debug_type"] == "DW_TAG_union_type":
            # skip for now
            return {}
        elif "struct" in t:
            combined_map = {}
            for offset in t["struct"]:
                member = t["struct"][offset]
                submap = get_pointer_map(member)
                for suboffset in submap:
                    combined_map[offset + suboffset] = submap[suboffset]
            return combined_map
        elif "subtype" in t:
            subtype = symbols_by_debug_address[t["type"]]
            pmap = get_pointer_map(subtype, depth + 1)
            size = get_size(subtype)
            expanded_map = {}
            for i in range(t["maxlen"]):
                for offset in pmap:
                    expanded_map[size * i + offset] = pmap[offset]
            return expanded_map
        else:
            print("no recurse", t)
            pass
        return {}

    # Do a second pass to dereference the types
    for symbol_address in symbols_by_memory_address:
        symbol = symbols_by_memory_address[symbol_address]
        if "type" in symbol:
            if symbol["debug_type"] == "DW_TAG_variable":
                symbol["pointer_map"] = get_pointer_map(symbols_by_debug_address[symbol["type"]])
            type_string = []
            t = symbol["type"]
            offset = []
            while t != None:
                t_symbol = symbols_by_debug_address[t]
                t = t_symbol.get("type", None)
                if "name" in t_symbol:
                    type_string.append(t_symbol["name"])
                elif t_symbol["debug_type"] == "DW_TAG_array_type":
                    type_string.append("[]")
                elif t_symbol["debug_type"] == "DW_TAG_pointer_type":
                    type_string.append("*")
                elif t_symbol["debug_type"] == "DW_TAG_const_type":
                    type_string.append("const")
                elif t_symbol["debug_type"] == "DW_TAG_volatile_type":
                    type_string.append("volatile")
                else:
                    # print("  ", t_symbol)
                    pass
            type_string.reverse()
            symbol["type_string"] = " ".join(type_string)
        # print(symbol_name, symbol["debug_type"], symbol.get("type_string", ""))

    # print()
    # print()
    # print(all_symbols["mp_builtin_module_table"])
    # return

    # Gather size and call info
    text_dump = objdump("-Dz", "-j", ".text", elf_filename)
    text_dump_lines = text_dump.stdout.decode("utf-8").split("\n")
    section = None
    symbol = None
    symbol_type = None
    for line in text_dump_lines[4:]:
        if line.startswith("Disassembly of section"):
            section = line.split()[-1].strip(":")
        elif not line:
            if symbol and "end_address" not in symbol:
                symbol["end_address"] = last_address
                symbol["size"] = last_address - symbol["start_address"]
                symbol = None
            continue
        elif line[0].isnumeric():
            symbol_address, symbol_name = line.split()
            symbol_address = parse_hex(symbol_address)
            symbol_name = symbol_name.strip("<>:")
            if symbol_name in symbols_by_linkage_name:
                linked_name = symbol_name
                symbol = symbols_by_linkage_name[symbol_name]
                if "name" in symbol:
                    non_linkage = symbol["name"]
                    if not non_linkage.startswith("__builtin"):
                        symbol_name = non_linkage
                all_symbols[symbol_name] = symbol
                if "name" not in symbol:
                    symbol["name"] = symbol_name
            elif symbol_address in symbols_by_memory_address:
                all_symbols[symbol_name] = symbols_by_memory_address[symbol_address]
                if "name" not in all_symbols[symbol_name]:
                    all_symbols[symbol_name]["name"] = symbol_name
            elif symbol_name not in all_symbols:
                if symbol_name == "nlr_push_tail_var":
                    fake_type = all_symbols["mp_obj_get_type"]["type"]
                    symbol = {
                        "debug_type": "DW_TAG_variable",
                        "name": symbol_name,
                        "type": fake_type,
                    }
                else:
                    print(line)
                    print(symbol_name, symbol_address)
                    symbol = {"debug_type": "DW_TAG_subprogram", "name": symbol_name}
                all_symbols[symbol_name] = symbol
                # raise RuntimeError()

            symbol = all_symbols[symbol_name]
            symbol["start_address"] = symbol_address
            symbols_by_memory_address[symbol_address] = symbol
            symbol["section"] = section

            if symbol["debug_type"] == "DW_TAG_subprogram":
                symbol["outgoing_jumps"] = set()
                symbol["incoming_jumps"] = set()
                symbol_type = None
            elif symbol["debug_type"] == "DW_TAG_variable":
                symbol["outgoing_pointers"] = set()
                symbol_type = symbols_by_debug_address[symbol["type"]]
            all_symbols[symbol_name] = symbol

        elif line[0] == " ":
            parts = line.strip().split()
            last_address = parse_hex(parts[0].strip(":"))

            offset = last_address - symbol["start_address"]
            if "pointer_map" in symbol:
                if offset not in symbol["pointer_map"]:
                    # print(offset, symbol)
                    pass
                else:
                    ref = parse_hex(parts[1])
                    pointer_style = symbol["pointer_map"][offset]
                    if pointer_style == MEMORY_POINTER:
                        symbol["outgoing_pointers"].add(ref & 0xFFFFFFFE)
                    elif pointer_style == MEMORY_PY_OBJECT and ref & 0x3 == 0:
                        symbol["outgoing_pointers"].add(ref)
            if len(parts[1]) == 8 and parts[1][0] == "0":
                addr = parse_hex(parts[1])
                if 0x2000 <= addr < 0x20000000:
                    if "outgoing_pointers" not in symbol:
                        symbol["outgoing_pointers"] = set()
                    symbol["outgoing_pointers"].add(addr)
            elif "<" in line and symbol["debug_type"] == "DW_TAG_subprogram":
                if line[-1] == ">":
                    jump_to = parts[-1].strip("<>").split("+")[0]
                    if "name" not in symbol:
                        print(jump_to)
                        print(symbol)
                    if jump_to != symbol["name"] and jump_to not in BAD_JUMPS:
                        symbol["outgoing_jumps"].add(jump_to)
                        # print(symbol_name, jump_to)
                        if jump_to == "_etext":
                            print(line)
                elif "UNDEFINED" in line:
                    continue
                elif parts[2] == "ldr":
                    continue
                else:
                    print(line)
        else:
            # print(line)
            pass

    # print()
    print(hex(min_call_site_param))
    print(all_symbols["exception_table"])
    # return

    print("converting outgoing pointers to names")

    # Convert outgoing pointers to names from addresses
    for symbol_name in all_symbols:
        symbol = all_symbols[symbol_name]
        if "outgoing_pointers" not in symbol:
            continue
        converted = set()
        for outgoing in symbol["outgoing_pointers"]:
            if outgoing in symbols_by_memory_address:
                outgoing = symbols_by_memory_address[outgoing]
                # print(outgoing)
                if outgoing["debug_type"] in ["DW_TAG_GNU_call_site", "DW_TAG_lexical_block"]:
                    continue
                if outgoing["name"] == "audioio_wavefile_type":
                    print(outgoing)
                converted.add(outgoing["name"])
        symbol["outgoing_pointers"] = converted

    print("linking back")
    # Link back
    for symbol_name in all_symbols:
        symbol = all_symbols[symbol_name]
        if "outgoing_jumps" in symbol:
            for outgoing in symbol["outgoing_jumps"]:
                if outgoing not in all_symbols:
                    # print(outgoing, symbol_name)
                    continue
                # print(all_symbols[outgoing], symbol_name)

                referenced_symbol = all_symbols[outgoing]
                if "incoming_jumps" not in referenced_symbol:
                    # print(symbol_name, "->", outgoing)
                    referenced_symbol["incoming_jumps"] = set()
                referenced_symbol["incoming_jumps"].add(symbol_name)
        if "outgoing_pointers" in symbol:
            for outgoing in symbol["outgoing_pointers"]:
                if outgoing not in all_symbols:
                    # print(outgoing, symbol_name)
                    continue
                # print(all_symbols[outgoing], symbol_name)

                referenced_symbol = all_symbols[outgoing]
                if "incoming_pointers" not in referenced_symbol:
                    # print(symbol_name, "->", outgoing)
                    referenced_symbol["incoming_pointers"] = set()
                referenced_symbol["incoming_pointers"].add(symbol_name)

    print(all_symbols["exception_table"])

    # Chart it all
    print("charting {} symbols".format(len(all_symbols)))
    callgraph = pgv.AGraph(directed=True)
    for i, symbol_name in enumerate(all_symbols):
        symbol = all_symbols[symbol_name]
        # print(i, symbol_name)
        # if "outgoing_jumps" in symbol:
        #     print("   ", len(symbol["outgoing_jumps"]), "jumps")
        # if "outgoing_pointers" in symbol:
        #     print("   ", len(symbol["outgoing_pointers"]), "ptrs")
        # if i > 3000:
        #     break
        if ("incoming_jumps" not in symbol or len(symbol["incoming_jumps"]) == 0) and (
            "incoming_pointers" not in symbol or len(symbol["incoming_pointers"]) == 0
        ):
            # print(symbol_name)
            continue
        if "start_address" not in symbol:
            continue
        callgraph.add_node(symbol_name)
        if "outgoing_jumps" in symbol:
            for outgoing in symbol["outgoing_jumps"]:
                callgraph.add_edge(symbol_name, outgoing)
        if "outgoing_pointers" in symbol:
            for outgoing in symbol["outgoing_pointers"]:
                callgraph.add_edge(symbol_name, outgoing, color="red")
        # print(symbol_name, symbol)

    # Style all of the nodes
    print("styling")
    for node in callgraph.iternodes():
        if node.name not in all_symbols:
            continue
        symbol = all_symbols[node.name]
        node.attr["shape"] = "box"
        text_width_ish = len(node.name) * 0.1
        if "size" not in symbol:
            print(symbol)
        size = symbol["size"] / 8
        square_size = size ** 0.5
        if text_width_ish > square_size:
            w = text_width_ish
            h = size / text_width_ish
        else:
            w = square_size
            h = square_size
        node.attr["width"] = w
        node.attr["height"] = h
        node.attr["label"] = node.name + "\r\n" + str(symbol["size"]) + " bytes"
        node.attr["style"] = "filled"

        incoming = 0
        if "incoming_jumps" in symbol:
            incoming += len(symbol["incoming_jumps"])
        if "incoming_pointers" in symbol:
            incoming += len(symbol["incoming_pointers"])

        if node.name in SPECIAL_NODE_COLORS:
            node.attr["color"] = SPECIAL_NODE_COLORS[node.name]
        elif incoming == 1:
            node.attr["color"] = "lightblue"
        elif incoming > 25:
            print("delete", node.name, "because it has {} incoming".format(incoming))
            callgraph.delete_node(node.name)
        elif incoming > 15:
            node.attr["color"] = "red"

    print("drawing")
    callgraph.layout(prog="dot")
    fn = "callgraph.svg"
    print(fn)
    callgraph.draw(fn)


if __name__ == "__main__":
    do_all_the_things()



# test math.isclose (appeared in Python 3.5)

try:
    from math import isclose
except ImportError:
    print("SKIP")
    raise SystemExit


def test(a, b, **kwargs):
    print(isclose(a, b, **kwargs))


def test_combinations(a, b, **kwargs):
    test(a, a, **kwargs)
    test(a, b, **kwargs)
    test(b, a, **kwargs)
    test(b, b, **kwargs)


# Special numbers
test_combinations(float("nan"), 1)
test_combinations(float("inf"), 1)
test_combinations(float("-inf"), 1)

# Equality
test(1.0, 1.0, rel_tol=0.0, abs_tol=0.0)
test(2.35e-100, 2.35e-100, rel_tol=0.0, abs_tol=0.0)
test(2.1234e100, 2.1234e100, rel_tol=0.0, abs_tol=0.0)

# Relative tolerance
test(1000.0, 1001.0, rel_tol=1e-3)
test(1000.0, 1001.0, rel_tol=1e-4)
test(1000, 1001, rel_tol=1e-3)
test(1000, 1001, rel_tol=1e-4)
test_combinations(0, 1, rel_tol=1.0)

# Absolute tolerance
test(0.0, 1e-10, abs_tol=1e-10, rel_tol=0.1)
test(0.0, 1e-10, abs_tol=0.0, rel_tol=0.1)

# Bad parameters
try:
    isclose(0, 0, abs_tol=-1)
except ValueError:
    print("ValueError")
try:
    isclose(0, 0, rel_tol=-1)
except ValueError:
    print("ValueError")



"""
This is a helper script for merging in new versions of MicroPython. You *must*
evaluate it's correctness and adapt it for each MP version. This is committed
in the repo more for reference than "fire and forget" use.
"""

import sh
from sh import git
from io import StringIO

out_buf = StringIO()

ports_to_delete = [
    "mimxrt",
    "powerpc",
    "samd",
    "javascript",
    "stm32",
    "esp32",
    "cc3200",
    "teensy",
    "windows",
    "zephyr",
    "minimal",
    "esp8266",
    "pic16bit",
    "qemu-arm",
    "bare-arm",
    "rp2",
]
for p in ports_to_delete:
    try:
        git.rm("-rf", "ports/" + p)
    except sh.ErrorReturnCode_128:
        pass

# We inherit stm32 changes into stm because we did a git rename.
git.status("--porcelain=1", "ports/stm", _out=out_buf)
out_buf.seek(0)
line = out_buf.readline()
while line:
    state, path = line.split()
    if state == "UU":
        git.checkout("--ours", path)
        git.add(path)
    elif state == "UA":
        git.rm(path)
    line = out_buf.readline()

# MicroPython added their nrf code in ports/nrf too. So, we always take our version.
out_buf = StringIO()
git.status("--porcelain=1", "ports/nrf", _out=out_buf)
out_buf.seek(0)
line = out_buf.readline()
while line:
    state, path = line.split()
    if state == "UU":
        git.checkout("--ours", path)
        git.add(path)
    elif state == "UA":
        git.rm(path)
    elif state == "AA":
        git.rm("-f", path)
    elif state == "A":
        git.rm("-f", path)
    elif state == "DU":
        git.rm(path)
    elif state == "DD":
        git.rm(path)
    else:
        print(state, path)
    line = out_buf.readline()


# MicroPython has their own CI settings. Let's not use them now.
out_buf = StringIO()
git.status("--porcelain=1", ".github/workflows", _out=out_buf)
out_buf.seek(0)
line = out_buf.readline()
while line:
    state, path = line.split()
    if state == "A":
        git.rm("-f", path)
    else:
        print(state, path)
    line = out_buf.readline()

# Delete docs and tests for things we don't need anymore
docs_to_delete = [
    "conf.py",
    "pyboard",
    "library/pyb.*",
    "library/esp*",
    "library/lcd160cr.rst",
    "esp8266",
    "wipy",
    "esp32",
    "library/machine.*",
    "library/network.*",
    "library/ubluetooth.rst",
    "library/ucryptolib.rst",
    "library/uos.rst",
    "library/usocket.rst",
    "library/ussl.rst",
    "library/ustruct.rst",
    "develop",
    "reference",
    "make.bat",
    "templates/topindex.html",
]
for d in docs_to_delete:
    try:
        git.rm("-rf", "docs/" + d)
    except sh.ErrorReturnCode_128:
        pass

tests_to_delete = [
    "wipy",
    "pyb",
    "multi_net",
    "net_inet",
    "multi_bluetooth",
    "esp32",
    "net_hosted",
]
for t in tests_to_delete:
    try:
        git.rm("-rf", "tests/" + t)
    except sh.ErrorReturnCode_128:
        pass


libs_to_delete = [
    "mynewt-nimble",
    "nxp_driver",
    "mbedtls",
    "mbedtls_errors",
    "asf4",
    "btstack",
]
for l in libs_to_delete:
    try:
        git.rm("-rf", "lib/" + l)
    except sh.ErrorReturnCode_128:
        pass

extmod_to_delete = [
    "machine_*",
    "webrepl",
    "uzlib",
    "ussl*",
    "modlwip.c",
    "moducryptolib.c",
    "modbluetooth*",
    "network*",
    "nimble",
    "btstack",
    "mpbthci*",
]
for e in extmod_to_delete:
    try:
        git.rm("-rf", "extmod/" + e)
    except sh.ErrorReturnCode_128 as error:
        print(error)
        pass

top_delete = [
    "drivers",
    ".travis.yml",
    ".github/FUNDING.yml",
    ".github/workflows/docs.yml",
    "README.md",
    "CODEOFCONDUCT.md",
    "CODECONVENTIONS.md",
    "examples",
]
for t in top_delete:
    try:
        git.rm("-rf", t)
    except sh.ErrorReturnCode_128:
        pass

# Always ours:
always_ours = [
    "devices",
    "supervisor",
    "shared-bindings",
    "shared-module",
    "ports/atmel-samd",
    "ports/cxd56",
    "ports/esp32s2",
    "ports/mimxrt10xx",
    "ports/raspberrypi",
    "ports/stm",
]
for ours in always_ours:
    out_buf = StringIO()
    git.status("--porcelain=1", ours, _out=out_buf)
    out_buf.seek(0)
    line = out_buf.readline()
    while line:
        state, path = line.split()
        if state == "UU":
            print("ours", path)
            git.checkout("--ours", path)
            git.add(path)
        else:
            print(state, path)
        line = out_buf.readline()

# Check to see if any files changed only in formatting
out_buf = StringIO()
git.status("--porcelain=1", ".", _out=out_buf)
out_buf.seek(0)
line = out_buf.readline()
while line:
    state = line.split()[0]
    if state in ("D", "R", "DD"):
        line = out_buf.readline()
        continue
    state, path = line.split()
    log_buf = StringIO()
    git.log("--pretty=tformat:%H", "25ae98f..HEAD", path, _out=log_buf, _tty_out=False)
    log_buf.seek(0)
    commits = []
    for line in log_buf.readlines():
        commits.append(line.strip())
    if state in ["UU", "M"] and commits == ["a52eb88031620a81521b937f2a0651dbac2bb350"]:
        git.checkout("--theirs", path)
        git.add(path)
    line = out_buf.readline()



#!/usr/bin/python3


def defines(name, suffix):
    print(f"mcu_pin_function_t {name} [] = {{")
    for instance in (0, 1):
        for function in "HI":
            for port in "ABCD":
                for idx in range(32):
                    pin = f"P{port}{idx:02d}"
                    pinmux = f"PINMUX_{pin}{function}_CAN{instance}_{suffix}"
                    print(
                        f"""\
#if defined({pinmux}) && ! defined(IGNORE_PIN_{pin})
    {{&pin_{pin}, {instance}, PIN_{pin}, {pinmux} & 0xffff}},
#endif"""
                    )
    print(f"{{NULL, 0, 0}}")
    print(f"}};")
    print()


print(
    """\
#include <stdint.h>
#include "py/obj.h"
#include "sam.h"
#include "samd/pins.h"
#include "mpconfigport.h"
#include "atmel_start_pins.h"
#include "hal/include/hal_gpio.h"
#include "common-hal/microcontroller/Pin.h"
"""
)

defines("can_rx", "RX")
defines("can_tx", "TX")



# This test requires CPython3.5

try:
    b'' % ()
except TypeError:
    print("SKIP")
    raise SystemExit

print(b"%%" % ())
print(b"=%d=" % 1)
print(b"=%d=%d=" % (1, 2))

print(b"=%s=" % b"str")
print(b"=%r=" % b"str")



class CtxMgr:
    def __init__(self, id):
        self.id = id

    def __enter__(self):
        print("__enter__", self.id)
        return self

    def __exit__(self, a, b, c):
        print("__exit__", self.id, repr(a), repr(b))

# simple case
def foo():
    with CtxMgr(1):
        return 4
print(foo())

# for loop within with (iterator needs removing upon return)
def f():
    with CtxMgr(1):
        for i in [1, 2]:
            return i
print(f())

# multiple for loops within with
def f():
    with CtxMgr(1):
        for i in [1, 2]:
            for j in [3, 4]:
                return (i, j)
print(f())

# multiple for loops within nested withs
def f():
    with CtxMgr(1):
        for i in [1, 2]:
            for j in [3, 4]:
                with CtxMgr(2):
                    for k in [5, 6]:
                        for l in [7, 8]:
                            return (i, j, k, l)
print(f())

# multiple for loops that are optimised, and nested withs
def f():
    with CtxMgr(1):
        for i in range(1, 3):
            for j in range(3, 5):
                with CtxMgr(2):
                    for k in range(5, 7):
                        for l in range(7, 9):
                            return (i, j, k, l)
print(f())



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import sys

output_filename = sys.argv[1]
input_filenames = {}
i = 2
while i < len(sys.argv):
    offset = int(sys.argv[i], 16)
    filename = sys.argv[i + 1]
    input_filenames[offset] = filename
    i += 2

with open(output_filename, "wb") as output_file:
    offsets = sorted(input_filenames.keys())
    for offset in offsets:
        input_filename = input_filenames[offset]
        # Fill with ones to save NOR flash.
        while output_file.tell() < offset:
            output_file.write(b"\xff")
        with open(input_filename, "rb") as input_file:
            output_file.write(input_file.read())



#! /usr/bin/env python3

# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import sys
import os
import json
import yaml

import build_board_info

workflow_file = ".github/workflows/build.yml"

# Get boards in json format
boards_info_json = build_board_info.get_board_mapping()

# Get all the boards out of the json format
info_boards = [
    board for board in boards_info_json.keys() if not boards_info_json[board].get("alias", False)
]

# We need to know the path of the workflow file
base_path = os.path.dirname(__file__)
yml_path = os.path.abspath(os.path.join(base_path, "..", workflow_file))

# Loading board list based on build jobs in the workflow file.
ci_boards = []
with open(yml_path, "r") as f:
    workflow = yaml.safe_load(f)

ok = True
for job in workflow["jobs"]:
    if not job.startswith("build"):
        continue
    job_boards = workflow["jobs"][job]["strategy"]["matrix"]["board"]
    if job_boards != sorted(job_boards):
        print('Boards for job "{}" not sorted. Must be:'.format(job))
        print('        - "' + '"\n        - "'.join(sorted(job_boards)) + '"')
        ok = False
    ci_boards.extend(job_boards)

# All the travis_boards elements must be on info_boards
info_boards.sort()
ci_boards.sort()

missing_boards = set(info_boards) - set(ci_boards)

if missing_boards:
    ok = False
    print("Boards missing in {}:".format(workflow_file))
    for board in missing_boards:
        print(board)

if not ok:
    sys.exit(1)



"""
categories: Syntax,Operators
description: MicroPython allows using := to assign to the variable of a comprehension, CPython raises a SyntaxError.
cause: MicroPython is optimised for code size and doesn't check this case.
workaround: Do not rely on this behaviour if writing CPython compatible code.
"""
print([i := -1 for i in range(4)])



"""
categories: Modules,array
description: Overflow checking is not implemented
cause: MicroPython implements implicit truncation in order to reduce code size and execution time
workaround: If CPython compatibility is needed then mask the value explicitly
"""
import array

a = array.array("b", [257])
print(a)



# Test current_task() function

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def task(result):
    result[0] = asyncio.current_task()


async def main():
    result = [None]
    t = asyncio.create_task(task(result))
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    print(t is result[0])


asyncio.run(main())



a = "¢пр"

print(a[0], a[0:1])
print(a[1], a[1:2])
print(a[2], a[2:3])
try:
    print(a[3])
except IndexError:
    print("IndexError")
print(a[3:4])

print(a[-1])
print(a[-2], a[-2:-1])
print(a[-3], a[-3:-2])
try:
    print(a[-4])
except IndexError:
    print("IndexError")
print(a[-4:-3])

print(a[0:2])
print(a[1:3])
print(a[2:4])



# test builtin property combined with inheritance
try:
    property
except:
    print("SKIP")
    raise SystemExit

# test property in a base class works for derived classes
class A:
    @property
    def x(self):
        print('A x')
        return 123
class B(A):
    pass
class C(B):
    pass
class D:
    pass
class E(C, D):
    pass
print(A().x)
print(B().x)
print(C().x)
print(E().x)

# test that we can add a property to base class after creation
class F:
    pass
F.foo = property(lambda self: print('foo get'))
class G(F):
    pass
F().foo
G().foo

# should be able to add a property to already-subclassed class because it already has one
F.bar = property(lambda self: print('bar get'))
F().bar
G().bar

# test case where class (H here) is already subclassed before adding attributes
class H:
    pass
class I(H):
    pass

# should be able to add a normal member to already-subclassed class
H.val = 2
print(I().val)

# should be able to add a property to the derived class
I.baz = property(lambda self: print('baz get'))
I().baz



"""
categories: Core,Generator
description: Context manager __exit__() not called in a generator which does not run to completion
cause: Unknown
workaround: Unknown
"""


class foo(object):
    def __enter__(self):
        print("Enter")

    def __exit__(self, *args):
        print("Exit")


def bar(x):
    with foo():
        while True:
            x += 1
            yield x


def func():
    g = bar(0)
    for _ in range(3):
        print(next(g))


func()



# test equality

print(None == None)

print(False == None)
print(False == False)
print(False == True)

print(() == ())
print(() == [])
print([] == [])
print(() == {})
print({} == ())

print(() == None)
print(() == False)
print(() == print)

print([] == None)
print([] == False)
print([] == print)

print({} == None)
print({} == False)
print({} == print)

print(1 == 1)
print(1 == 2)
print(1 == ())
print(1 == [])
print(1 == {})
print(1 == 'a')

print('a' == 'a')
print('a' == 'ab')
print('a' == 1)
print('a' == ())

# same as above but with !=

print(None != None)

print(False != None)
print(False != False)
print(False != True)

print(() != ())
print(() != [])
print([] != [])
print(() != {})
print({} != ())

print(() != None)
print(() != False)
print(() != print)

print([] != None)
print([] != False)
print([] != print)

print({} != None)
print({} != False)
print({} != print)

print(1 != 1)
print(1 != 2)
print(1 != ())
print(1 != [])
print(1 != {})
print(1 != 'a')

print('a' != 'a')
print('a' != 'ab')
print('a' != 1)
print('a' != ())



# test VFS functionality without any particular filesystem type

try:
    import uos

    uos.mount
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class Filesystem:
    def __init__(self, id, fail=0):
        self.id = id
        self.fail = fail

    def mount(self, readonly, mkfs):
        print(self.id, "mount", readonly, mkfs)

    def umount(self):
        print(self.id, "umount")

    def ilistdir(self, dir):
        print(self.id, "ilistdir", dir)
        return iter([("a%d" % self.id, 0, 0)])

    def chdir(self, dir):
        print(self.id, "chdir", dir)
        if self.fail:
            raise OSError(self.fail)

    def getcwd(self):
        print(self.id, "getcwd")
        return "dir%d" % self.id

    def mkdir(self, path):
        print(self.id, "mkdir", path)

    def remove(self, path):
        print(self.id, "remove", path)

    def rename(self, old_path, new_path):
        print(self.id, "rename", old_path, new_path)

    def rmdir(self, path):
        print(self.id, "rmdir", path)

    def stat(self, path):
        print(self.id, "stat", path)
        return (self.id,)

    def statvfs(self, path):
        print(self.id, "statvfs", path)
        return (self.id,)

    def open(self, file, mode):
        print(self.id, "open", file, mode)


# first we umount any existing mount points the target may have
try:
    uos.umount("/")
except OSError:
    pass
for path in uos.listdir("/"):
    uos.umount("/" + path)

# stat root dir
print(uos.stat("/"))

# statvfs root dir; verify that f_namemax has a sensible size
print(uos.statvfs("/")[9] >= 32)

# getcwd when in root dir
print(uos.getcwd())

# test operations on the root directory with nothing mounted, they should all fail
for func in ("chdir", "listdir", "mkdir", "remove", "rmdir", "stat"):
    for arg in ("x", "/x"):
        try:
            getattr(uos, func)(arg)
        except OSError:
            print(func, arg, "OSError")

# basic mounting and listdir
uos.mount(Filesystem(1), "/test_mnt")
print(uos.listdir())

# ilistdir
i = uos.ilistdir()
print(next(i))
try:
    next(i)
except StopIteration:
    print("StopIteration")
try:
    next(i)
except StopIteration:
    print("StopIteration")

# referencing the mount point in different ways
print(uos.listdir("test_mnt"))
print(uos.listdir("/test_mnt"))

# mounting another filesystem
uos.mount(Filesystem(2), "/test_mnt2", readonly=True)
print(uos.listdir())
print(uos.listdir("/test_mnt2"))

# mounting over an existing mount point
try:
    uos.mount(Filesystem(3), "/test_mnt2")
except OSError:
    print("OSError")

# mkdir of a mount point
try:
    uos.mkdir("/test_mnt")
except OSError:
    print("OSError")

# rename across a filesystem
try:
    uos.rename("/test_mnt/a", "/test_mnt2/b")
except OSError:
    print("OSError")

# delegating to mounted filesystem
uos.chdir("test_mnt")
print(uos.listdir())
print(uos.getcwd())
uos.mkdir("test_dir")
uos.remove("test_file")
uos.rename("test_file", "test_file2")
uos.rmdir("test_dir")
print(uos.stat("test_file"))
print(uos.statvfs("/test_mnt"))
open("test_file")
open("test_file", "wb")

# umount
uos.umount("/test_mnt")
uos.umount("/test_mnt2")

# umount a non-existent mount point
try:
    uos.umount("/test_mnt")
except OSError:
    print("OSError")

# root dir
uos.mount(Filesystem(3), "/")
print(uos.stat("/"))
print(uos.statvfs("/"))
print(uos.listdir())
open("test")

uos.mount(Filesystem(4), "/mnt")
print(uos.listdir())
print(uos.listdir("/mnt"))
uos.chdir("/mnt")
print(uos.listdir())

# chdir to a subdir within root-mounted vfs, and then listdir
uos.chdir("/subdir")
print(uos.listdir())
uos.chdir("/")

uos.umount("/")
print(uos.listdir("/"))
uos.umount("/mnt")

# chdir to a non-existent mount point (current directory should remain unchanged)
try:
    uos.chdir("/foo")
except OSError:
    print("OSError")
print(uos.getcwd())

# chdir to a non-existent subdirectory in a mounted filesystem
uos.mount(Filesystem(5, 1), "/mnt")
try:
    uos.chdir("/mnt/subdir")
except OSError:
    print("OSError")
print(uos.getcwd())



# test vsqrt, vneg
@micropython.asm_thumb  # r0 = -(int)(sqrt(r0)*r1)
def sqrt_test(r0, r1):
    vmov(s1, r0)
    vcvt_f32_s32(s1, s1)
    vsqrt(s1, s1)
    vmov(s2, r1)
    vcvt_f32_s32(s2, s2)
    vmul(s0, s1, s2)
    vneg(s7, s0)
    vcvt_s32_f32(s31, s7)
    vmov(r0, s31)


print(sqrt_test(256, 10))



try:
    # If we don't expose object.__new__ (small ports), there's
    # nothing to test.
    object.__new__
except AttributeError:
    print("SKIP")
    raise SystemExit

class A:
    def __new__(cls):
        print("A.__new__")
        return super(cls, A).__new__(cls)

    def __init__(self):
        print("A.__init__")

    def meth(self):
        print('A.meth')

#print(A.__new__)
#print(A.__init__)

a = A()
a.meth()

a = A.__new__(A)
a.meth()

#print(a.meth)
#print(a.__init__)
#print(a.__new__)

# __new__ should automatically be a staticmethod, so this should work
a = a.__new__(A)
a.meth()

# __new__ returns not an instance of the class (None here), __init__
# should not be called

class B:
    def __new__(self, v1, v2):
        print("B.__new__", v1, v2)

    def __init__(self, v1, v2):
        # Should not be called in this test
        print("B.__init__", v1, v2)

print("B inst:", B(1, 2))


# Variation of the above, __new__ returns an instance of another class,
# __init__ should not be called

class Dummy: pass

class C:
    def __new__(cls):
        print("C.__new__")
        return Dummy()

    def __init__(self):
        # Should not be called in this test
        print("C.__init__")

c = C()
print(isinstance(c, Dummy))



# Make sure that write operations on io.BytesIO don't
# change original object it was constructed from.
try:
    import uio as io
except ImportError:
    import io

b = b"foobar"

a = io.BytesIO(b)
a.write(b"1")
print(b)
print(a.getvalue())

b = bytearray(b"foobar")

a = io.BytesIO(b)
a.write(b"1")
print(b)
print(a.getvalue())



# tests for natively compiled functions

# basic test
@micropython.native
def native_test(x):
    print(1, [], x)


native_test(2)

# check that GC doesn't collect the native function
import gc

gc.collect()
native_test(3)

# native with 2 args
@micropython.native
def f(a, b):
    print(a + b)


f(1, 2)

# native with 3 args
@micropython.native
def f(a, b, c):
    print(a + b + c)


f(1, 2, 3)

# check not operator
@micropython.native
def f(a):
    print(not a)


f(False)
f(True)



# test user defined iterators

# this class is not iterable
class NotIterable:
    pass
try:
    for i in NotIterable():
        pass
except TypeError:
    print('TypeError')

# this class has no __next__ implementation
class NotIterable:
    def __iter__(self):
        return self
try:
    print(all(NotIterable()))
except TypeError:
    print('TypeError')

class MyStopIteration(StopIteration):
    pass

class myiter:
    def __init__(self, i):
        self.i = i

    def __iter__(self):
        return self

    def __next__(self):
        if self.i <= 0:
            # stop in the usual way
            raise StopIteration
        elif self.i == 10:
            # stop with an argument
            raise StopIteration(42)
        elif self.i == 20:
            # raise a non-stop exception
            raise TypeError
        elif self.i == 30:
            # raise a user-defined stop iteration
            print('raising MyStopIteration')
            raise MyStopIteration
        else:
            # return the next value
            self.i -= 1
            return self.i

for i in myiter(5):
    print(i)

for i in myiter(12):
    print(i)

try:
    for i in myiter(22):
        print(i)
except TypeError:
    print('raised TypeError')

try:
    for i in myiter(5):
        print(i)
        raise StopIteration
except StopIteration:
    print('raised StopIteration')

for i in myiter(32):
    print(i)

# repeat some of the above tests but use tuple() to walk the iterator (tests mp_iternext)
print(tuple(myiter(5)))
print(tuple(myiter(12)))
print(tuple(myiter(32)))
try:
    tuple(myiter(22))
except TypeError:
    print('raised TypeError')



# test round() with floats

# check basic cases
tests = [
    [0.0],
    [1.0],
    [0.1],
    [-0.1],
    [123.4],
    [123.6],
    [-123.4],
    [-123.6],
    [1.234567, 5],
    [1.23456, 1],
    [1.23456, 0],
    [1234.56, -2],
]
for t in tests:
    print(round(*t))

# check .5 cases
for i in range(11):
    print(round((i - 5) / 2))

# test second arg
for i in range(-1, 3):
    print(round(1.47, i))

# test inf and nan
for val in (float("inf"), float("nan")):
    try:
        round(val)
    except (ValueError, OverflowError) as e:
        print(type(e))



class foo(object):
    def __init__(self, value):
        self.x = value

    def __eq__(self, other):
        print('eq')
        return self.x == other.x

    def __lt__(self, other):
        print('lt')
        return self.x < other.x

    def __gt__(self, other):
        print('gt')
        return self.x > other.x

    def __le__(self, other):
        print('le')
        return self.x <= other.x

    def __ge__(self, other):
        print('ge')
        return self.x >= other.x

for i in range(3):
    for j in range(3):
        print(foo(i) == foo(j))
        print(foo(i) < foo(j))
        print(foo(i) > foo(j))
        print(foo(i) <= foo(j))
        print(foo(i) >= foo(j))



# Array operation
# Type: list, map() call. This method requires allocation of
# the same amount of memory as original array (to hold result
# array). On the other hand, input array stays intact.
import bench


def test(num):
    for i in iter(range(num // 10000)):
        arr = bytearray(b"\0" * 1000)
        arr2 = bytearray(map(lambda x: x + 1, arr))


bench.run(test)



# test syntax errors using async

try:
    exec
except NameError:
    print("SKIP")
    raise SystemExit


def test_syntax(code):
    try:
        exec(code)
        print("no SyntaxError")
    except SyntaxError:
        print("SyntaxError")


test_syntax("async for x in (): x")
test_syntax("async with x: x")



# test _thread lock object using a single thread
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import _thread

# create lock
lock = _thread.allocate_lock()

print(type(lock) == _thread.LockType)

# should be unlocked
print(lock.locked())

# basic acquire and release
print(lock.acquire())
print(lock.locked())
lock.release()
print(lock.locked())

# try acquire twice (second should fail)
print(lock.acquire())
print(lock.locked())
print(lock.acquire(0))
print(lock.locked())
lock.release()
print(lock.locked())

# test with capabilities of lock
with lock:
    print(lock.locked())

# test that lock is unlocked if an error is rasied
try:
    with lock:
        print(lock.locked())
        raise KeyError
except KeyError:
    print("KeyError")
    print(lock.locked())

# test that we can't release an unlocked lock
try:
    lock.release()
except RuntimeError:
    print("RuntimeError")



b = bytearray(30)
f = open("io/data/file1", "rb")
# 2nd arg (length to read) is extension to CPython
print(f.readinto(b, 8))
print(b)

b = bytearray(4)
f = open("io/data/file1", "rb")
print(f.readinto(b, 8))
print(b)



# test bl and bx instructions


@micropython.asm_thumb
def f(r0):
    # jump over the internal functions
    b(entry)

    label(func1)
    add(r0, 2)
    bx(lr)

    label(func2)
    sub(r0, 1)
    bx(lr)

    label(entry)
    bl(func1)
    bl(func2)


print(f(0))
print(f(1))



# test attrtuple
# we can't test this type directly so we use sys.implementation object

import sys
t = sys.implementation

# It can be just a normal tuple on small ports
try:
    t.name
except AttributeError:
    print("SKIP")
    raise SystemExit


# test printing of attrtuple
print(str(t).find("version=") > 0)

# test read attr
print(isinstance(t.name, str))



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

"""
Generate documentation for pyboard API from C files.
"""

import os
import argparse
import re
import markdown

# given a list of (name,regex) pairs, find the first one that matches the given line
def re_match_first(regexs, line):
    for name, regex in regexs:
        match = re.match(regex, line)
        if match:
            return name, match
    return None, None


def makedirs(d):
    if not os.path.isdir(d):
        os.makedirs(d)


class Lexer:
    class LexerError(Exception):
        pass

    class EOF(Exception):
        pass

    class Break(Exception):
        pass

    def __init__(self, file):
        self.filename = file
        with open(file, "rt") as f:
            line_num = 0
            lines = []
            for line in f:
                line_num += 1
                line = line.strip()
                if line == "///":
                    lines.append((line_num, ""))
                elif line.startswith("/// "):
                    lines.append((line_num, line[4:]))
                elif len(lines) > 0 and lines[-1][1] is not None:
                    lines.append((line_num, None))
        if len(lines) > 0 and lines[-1][1] is not None:
            lines.append((line_num, None))
        self.cur_line = 0
        self.lines = lines

    def opt_break(self):
        if len(self.lines) > 0 and self.lines[0][1] is None:
            self.lines.pop(0)

    def next(self):
        if len(self.lines) == 0:
            raise Lexer.EOF
        else:
            l = self.lines.pop(0)
            self.cur_line = l[0]
            if l[1] is None:
                raise Lexer.Break
            else:
                return l[1]

    def error(self, msg):
        print("({}:{}) {}".format(self.filename, self.cur_line, msg))
        raise Lexer.LexerError


class MarkdownWriter:
    def __init__(self):
        pass

    def start(self):
        self.lines = []

    def end(self):
        return "\n".join(self.lines)

    def heading(self, level, text):
        if len(self.lines) > 0:
            self.lines.append("")
        self.lines.append(level * "#" + " " + text)
        self.lines.append("")

    def para(self, text):
        if len(self.lines) > 0 and self.lines[-1] != "":
            self.lines.append("")
        if isinstance(text, list):
            self.lines.extend(text)
        elif isinstance(text, str):
            self.lines.append(text)
        else:
            assert False
        self.lines.append("")

    def single_line(self, text):
        self.lines.append(text)

    def module(self, name, short_descr, descr):
        self.heading(1, "module {}".format(name))
        self.para(descr)

    def function(self, ctx, name, args, descr):
        proto = "{}.{}{}".format(ctx, self.name, self.args)
        self.heading(3, "`" + proto + "`")
        self.para(descr)

    def method(self, ctx, name, args, descr):
        if name == "\\constructor":
            proto = "{}{}".format(ctx, args)
        elif name == "\\call":
            proto = "{}{}".format(ctx, args)
        else:
            proto = "{}.{}{}".format(ctx, name, args)
        self.heading(3, "`" + proto + "`")
        self.para(descr)

    def constant(self, ctx, name, descr):
        self.single_line("`{}.{}` - {}".format(ctx, name, descr))


class ReStructuredTextWriter:
    head_chars = {1: "=", 2: "-", 3: "."}

    def __init__(self):
        pass

    def start(self):
        self.lines = []

    def end(self):
        return "\n".join(self.lines)

    def _convert(self, text):
        return text.replace("`", "``").replace("*", "\\*")

    def heading(self, level, text, convert=True):
        if len(self.lines) > 0:
            self.lines.append("")
        if convert:
            text = self._convert(text)
        self.lines.append(text)
        self.lines.append(len(text) * self.head_chars[level])
        self.lines.append("")

    def para(self, text, indent=""):
        if len(self.lines) > 0 and self.lines[-1] != "":
            self.lines.append("")
        if isinstance(text, list):
            for t in text:
                self.lines.append(indent + self._convert(t))
        elif isinstance(text, str):
            self.lines.append(indent + self._convert(text))
        else:
            assert False
        self.lines.append("")

    def single_line(self, text):
        self.lines.append(self._convert(text))

    def module(self, name, short_descr, descr):
        self.heading(1, ":mod:`{}` --- {}".format(name, self._convert(short_descr)), convert=False)
        self.lines.append(".. module:: {}".format(name))
        self.lines.append("   :synopsis: {}".format(short_descr))
        self.para(descr)

    def function(self, ctx, name, args, descr):
        args = self._convert(args)
        self.lines.append(".. function:: " + name + args)
        self.para(descr, indent="   ")

    def method(self, ctx, name, args, descr):
        args = self._convert(args)
        if name == "\\constructor":
            self.lines.append(".. class:: " + ctx + args)
        elif name == "\\call":
            self.lines.append(".. method:: " + ctx + args)
        else:
            self.lines.append(".. method:: " + ctx + "." + name + args)
        self.para(descr, indent="   ")

    def constant(self, ctx, name, descr):
        self.lines.append(".. data:: " + name)
        self.para(descr, indent="   ")


class DocValidateError(Exception):
    pass


class DocItem:
    def __init__(self):
        self.doc = []

    def add_doc(self, lex):
        try:
            while True:
                line = lex.next()
                if len(line) > 0 or len(self.doc) > 0:
                    self.doc.append(line)
        except Lexer.Break:
            pass

    def dump(self, writer):
        writer.para(self.doc)


class DocConstant(DocItem):
    def __init__(self, name, descr):
        super().__init__()
        self.name = name
        self.descr = descr

    def dump(self, ctx, writer):
        writer.constant(ctx, self.name, self.descr)


class DocFunction(DocItem):
    def __init__(self, name, args):
        super().__init__()
        self.name = name
        self.args = args

    def dump(self, ctx, writer):
        writer.function(ctx, self.name, self.args, self.doc)


class DocMethod(DocItem):
    def __init__(self, name, args):
        super().__init__()
        self.name = name
        self.args = args

    def dump(self, ctx, writer):
        writer.method(ctx, self.name, self.args, self.doc)


class DocClass(DocItem):
    def __init__(self, name, descr):
        super().__init__()
        self.name = name
        self.descr = descr
        self.constructors = {}
        self.classmethods = {}
        self.methods = {}
        self.constants = {}

    def process_classmethod(self, lex, d):
        name = d["id"]
        if name == "\\constructor":
            dict_ = self.constructors
        else:
            dict_ = self.classmethods
        if name in dict_:
            lex.error("multiple definition of method '{}'".format(name))
        method = dict_[name] = DocMethod(name, d["args"])
        method.add_doc(lex)

    def process_method(self, lex, d):
        name = d["id"]
        dict_ = self.methods
        if name in dict_:
            lex.error("multiple definition of method '{}'".format(name))
        method = dict_[name] = DocMethod(name, d["args"])
        method.add_doc(lex)

    def process_constant(self, lex, d):
        name = d["id"]
        if name in self.constants:
            lex.error("multiple definition of constant '{}'".format(name))
        self.constants[name] = DocConstant(name, d["descr"])
        lex.opt_break()

    def dump(self, writer):
        writer.heading(1, "class {}".format(self.name))
        super().dump(writer)
        if len(self.constructors) > 0:
            writer.heading(2, "Constructors")
            for f in sorted(self.constructors.values(), key=lambda x: x.name):
                f.dump(self.name, writer)
        if len(self.classmethods) > 0:
            writer.heading(2, "Class methods")
            for f in sorted(self.classmethods.values(), key=lambda x: x.name):
                f.dump(self.name, writer)
        if len(self.methods) > 0:
            writer.heading(2, "Methods")
            for f in sorted(self.methods.values(), key=lambda x: x.name):
                f.dump(self.name.lower(), writer)
        if len(self.constants) > 0:
            writer.heading(2, "Constants")
            for c in sorted(self.constants.values(), key=lambda x: x.name):
                c.dump(self.name, writer)


class DocModule(DocItem):
    def __init__(self, name, descr):
        super().__init__()
        self.name = name
        self.descr = descr
        self.functions = {}
        self.constants = {}
        self.classes = {}
        self.cur_class = None

    def new_file(self):
        self.cur_class = None

    def process_function(self, lex, d):
        name = d["id"]
        if name in self.functions:
            lex.error("multiple definition of function '{}'".format(name))
        function = self.functions[name] = DocFunction(name, d["args"])
        function.add_doc(lex)

    # def process_classref(self, lex, d):
    #    name = d['id']
    #    self.classes[name] = name
    #    lex.opt_break()

    def process_class(self, lex, d):
        name = d["id"]
        if name in self.classes:
            lex.error("multiple definition of class '{}'".format(name))
        self.cur_class = self.classes[name] = DocClass(name, d["descr"])
        self.cur_class.add_doc(lex)

    def process_classmethod(self, lex, d):
        self.cur_class.process_classmethod(lex, d)

    def process_method(self, lex, d):
        self.cur_class.process_method(lex, d)

    def process_constant(self, lex, d):
        if self.cur_class is None:
            # a module-level constant
            name = d["id"]
            if name in self.constants:
                lex.error("multiple definition of constant '{}'".format(name))
            self.constants[name] = DocConstant(name, d["descr"])
            lex.opt_break()
        else:
            # a class-level constant
            self.cur_class.process_constant(lex, d)

    def validate(self):
        if self.descr is None:
            raise DocValidateError("module {} referenced but never defined".format(self.name))

    def dump(self, writer):
        writer.module(self.name, self.descr, self.doc)
        if self.functions:
            writer.heading(2, "Functions")
            for f in sorted(self.functions.values(), key=lambda x: x.name):
                f.dump(self.name, writer)
        if self.constants:
            writer.heading(2, "Constants")
            for c in sorted(self.constants.values(), key=lambda x: x.name):
                c.dump(self.name, writer)
        if self.classes:
            writer.heading(2, "Classes")
            for c in sorted(self.classes.values(), key=lambda x: x.name):
                writer.para("[`{}.{}`]({}) - {}".format(self.name, c.name, c.name, c.descr))

    def write_html(self, dir):
        md_writer = MarkdownWriter()
        md_writer.start()
        self.dump(md_writer)
        with open(os.path.join(dir, "index.html"), "wt") as f:
            f.write(markdown.markdown(md_writer.end()))
        for c in self.classes.values():
            class_dir = os.path.join(dir, c.name)
            makedirs(class_dir)
            md_writer.start()
            md_writer.para("part of the [{} module](./)".format(self.name))
            c.dump(md_writer)
            with open(os.path.join(class_dir, "index.html"), "wt") as f:
                f.write(markdown.markdown(md_writer.end()))

    def write_rst(self, dir):
        rst_writer = ReStructuredTextWriter()
        rst_writer.start()
        self.dump(rst_writer)
        with open(dir + "/" + self.name + ".rst", "wt") as f:
            f.write(rst_writer.end())
        for c in self.classes.values():
            rst_writer.start()
            c.dump(rst_writer)
            with open(dir + "/" + self.name + "." + c.name + ".rst", "wt") as f:
                f.write(rst_writer.end())


class Doc:
    def __init__(self):
        self.modules = {}
        self.cur_module = None

    def new_file(self):
        self.cur_module = None
        for m in self.modules.values():
            m.new_file()

    def check_module(self, lex):
        if self.cur_module is None:
            lex.error("module not defined")

    def process_module(self, lex, d):
        name = d["id"]
        if name not in self.modules:
            self.modules[name] = DocModule(name, None)
        self.cur_module = self.modules[name]
        if self.cur_module.descr is not None:
            lex.error("multiple definition of module '{}'".format(name))
        self.cur_module.descr = d["descr"]
        self.cur_module.add_doc(lex)

    def process_moduleref(self, lex, d):
        name = d["id"]
        if name not in self.modules:
            self.modules[name] = DocModule(name, None)
        self.cur_module = self.modules[name]
        lex.opt_break()

    def process_class(self, lex, d):
        self.check_module(lex)
        self.cur_module.process_class(lex, d)

    def process_function(self, lex, d):
        self.check_module(lex)
        self.cur_module.process_function(lex, d)

    def process_classmethod(self, lex, d):
        self.check_module(lex)
        self.cur_module.process_classmethod(lex, d)

    def process_method(self, lex, d):
        self.check_module(lex)
        self.cur_module.process_method(lex, d)

    def process_constant(self, lex, d):
        self.check_module(lex)
        self.cur_module.process_constant(lex, d)

    def validate(self):
        for m in self.modules.values():
            m.validate()

    def dump(self, writer):
        writer.heading(1, "Modules")
        writer.para("These are the Python modules that are implemented.")
        for m in sorted(self.modules.values(), key=lambda x: x.name):
            writer.para("[`{}`]({}/) - {}".format(m.name, m.name, m.descr))

    def write_html(self, dir):
        md_writer = MarkdownWriter()
        with open(os.path.join(dir, "module", "index.html"), "wt") as f:
            md_writer.start()
            self.dump(md_writer)
            f.write(markdown.markdown(md_writer.end()))
        for m in self.modules.values():
            mod_dir = os.path.join(dir, "module", m.name)
            makedirs(mod_dir)
            m.write_html(mod_dir)

    def write_rst(self, dir):
        # with open(os.path.join(dir, 'module', 'index.html'), 'wt') as f:
        #    f.write(markdown.markdown(self.dump()))
        for m in self.modules.values():
            m.write_rst(dir)


regex_descr = r"(?P<descr>.*)"

doc_regexs = (
    (Doc.process_module, re.compile(r"\\module (?P<id>[a-z][a-z0-9]*) - " + regex_descr + r"$")),
    (Doc.process_moduleref, re.compile(r"\\moduleref (?P<id>[a-z]+)$")),
    (Doc.process_function, re.compile(r"\\function (?P<id>[a-z0-9_]+)(?P<args>\(.*\))$")),
    (Doc.process_classmethod, re.compile(r"\\classmethod (?P<id>\\?[a-z0-9_]+)(?P<args>\(.*\))$")),
    (Doc.process_method, re.compile(r"\\method (?P<id>\\?[a-z0-9_]+)(?P<args>\(.*\))$")),
    (
        Doc.process_constant,
        re.compile(r"\\constant (?P<id>[A-Za-z0-9_]+) - " + regex_descr + r"$"),
    ),
    # (Doc.process_classref, re.compile(r'\\classref (?P<id>[A-Za-z0-9_]+)$')),
    (Doc.process_class, re.compile(r"\\class (?P<id>[A-Za-z0-9_]+) - " + regex_descr + r"$")),
)


def process_file(file, doc):
    lex = Lexer(file)
    doc.new_file()
    try:
        try:
            while True:
                line = lex.next()
                fun, match = re_match_first(doc_regexs, line)
                if fun == None:
                    lex.error("unknown line format: {}".format(line))
                fun(doc, lex, match.groupdict())

        except Lexer.Break:
            lex.error("unexpected break")

        except Lexer.EOF:
            pass

    except Lexer.LexerError:
        return False

    return True


def main():
    cmd_parser = argparse.ArgumentParser(
        description="Generate documentation for pyboard API from C files."
    )
    cmd_parser.add_argument(
        "--outdir", metavar="<output dir>", default="gendoc-out", help="ouput directory"
    )
    cmd_parser.add_argument("--format", default="html", help="output format: html or rst")
    cmd_parser.add_argument("files", nargs="+", help="input files")
    args = cmd_parser.parse_args()

    doc = Doc()
    for file in args.files:
        print("processing", file)
        if not process_file(file, doc):
            return
    try:
        doc.validate()
    except DocValidateError as e:
        print(e)

    makedirs(args.outdir)

    if args.format == "html":
        doc.write_html(args.outdir)
    elif args.format == "rst":
        doc.write_rst(args.outdir)
    else:
        print("unknown format:", args.format)
        return

    print("written to", args.outdir)


if __name__ == "__main__":
    main()



"""
categories: Types,Exception
description: Exception in while loop condition may have unexpected line number
cause: Condition checks are optimized to happen at the end of loop body, and that line number is reported.
workaround: Unknown
"""
l = ["-foo", "-bar"]

i = 0
while l[i][0] == "-":
    print("iter")
    i += 1



def f():
    n = 0
    while True:
        n = yield n + 1
        print(n)

g = f()
try:
    g.send(1)
except TypeError:
    print("caught")

print(g.send(None))
print(g.send(100))
print(g.send(200))


def f2():
    print("entering")
    for i in range(3):
        print(i)
        yield
    print("returning 1")
    print("returning 2")

g = f2()
g.send(None)
g.send(1)
g.send(1)
try:
    g.send(1)
except StopIteration:
    print("caught")
try:
    g.send(1)
except StopIteration:
    print("caught")



print("foobar".startswith("foo"))
print("foobar".startswith("Foo"))
print("foobar".startswith("foo1"))
print("foobar".startswith("foobar"))
print("foobar".startswith(""))

print("1foobar".startswith("foo", 1))
print("1foo".startswith("foo", 1))
print("1foo".startswith("1foo", 1))
print("1fo".startswith("foo", 1))
print("1fo".startswith("foo", 10))

try:
    "foobar".startswith(1)
except TypeError:
    print("TypeError")



try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


try:
    import utime

    ticks = utime.ticks_ms
    ticks_diff = utime.ticks_diff
except:
    import time

    ticks = lambda: int(time.time() * 1000)
    ticks_diff = lambda t1, t0: t1 - t0


async def delay_print(t, s):
    await asyncio.sleep(t)
    print(s)


async def main():
    print("start")

    await asyncio.sleep(0.001)
    print("after sleep")

    t0 = ticks()
    await delay_print(0.02, "short")
    t1 = ticks()
    await delay_print(0.04, "long")
    t2 = ticks()
    await delay_print(-1, "negative")
    t3 = ticks()

    print(
        "took {} {} {}".format(
            round(ticks_diff(t1, t0), -1),
            round(ticks_diff(t2, t1), -1),
            round(ticks_diff(t3, t2), -1),
        )
    )


asyncio.run(main())



try:
    from uio import StringIO
    import ujson as json
except:
    try:
        from io import StringIO
        import json
    except ImportError:
        print("SKIP")
        raise SystemExit

print(json.load(StringIO("null")))
print(json.load(StringIO('"abc\\u0064e"')))
print(json.load(StringIO("[false, true, 1, -2]")))
print(json.load(StringIO('{"a":true}')))



# basic for loop

def f():
    for x in range(2):
        for y in range(2):
            for z in range(2):
                print(x, y, z)

f()

# range with negative step
for i in range(3, -1, -1):
    print(i)

a = -1
# range with non-constant step - we optimize constant steps, so this
# will be executed differently
for i in range(3, -1, a):
    print(i)



# test handling of failed heap allocation with tuple

import micropython

# create tuple
x = 1
micropython.heap_lock()
try:
    (x,)
except MemoryError:
    print("MemoryError: tuple create")
micropython.heap_unlock()



try:
    frozenset
except NameError:
    print("SKIP")
    raise SystemExit

l = [1, 2, 3, 4]
s = frozenset(l)
outs = [s.difference(),
        s.difference(frozenset({1})),
        s.difference(frozenset({1}), [1, 2]),
        s.difference(frozenset({1}), {1, 2}, {2, 3})]
for out in outs:
    print(type(out), sorted(out))

s = frozenset(l)
try:
    print(s.difference_update({1}))
except AttributeError:
    print("AttributeError")



# Change the following to True to get a much more comprehensive set of tests
# to run, albeit, which take considerably longer.

full_tests = False


def test(fmt, *args):
    print("{:8s}".format(fmt) + ">" + fmt.format(*args) + "<")


def test_fmt(conv, fill, alignment, sign, prefix, width, precision, type, arg):
    fmt = "{"
    if conv:
        fmt += "!"
        fmt += conv
    fmt += ":"
    if alignment:
        fmt += fill
        fmt += alignment
    fmt += sign
    fmt += prefix
    fmt += width
    if precision:
        fmt += "."
        fmt += precision
    fmt += type
    fmt += "}"
    test(fmt, arg)
    if fill == "0" and alignment == "=":
        fmt = "{:"
        fmt += sign
        fmt += prefix
        fmt += width
        if precision:
            fmt += "."
            fmt += precision
        fmt += type
        fmt += "}"
        test(fmt, arg)


eg_nums = (
    0.0,
    -0.0,
    0.1,
    1.234,
    12.3459,
    1.23456789,
    123456789.0,
    -0.0,
    -0.1,
    -1.234,
    -12.3459,
    1e4,
    1e-4,
    1e5,
    1e-5,
    1e6,
    1e-6,
    1e10,
    1e37,
    -1e37,
    1e-37,
    -1e-37,
    1.23456e8,
    1.23456e7,
    1.23456e6,
    1.23456e5,
    1.23456e4,
    1.23456e3,
    1.23456e2,
    1.23456e1,
    1.23456e0,
    1.23456e-1,
    1.23456e-2,
    1.23456e-3,
    1.23456e-4,
    1.23456e-5,
    1.23456e-6,
    1.23456e-7,
    1.23456e-8,
    -1.23456e8,
    -1.23456e7,
    -1.23456e6,
    -1.23456e5,
    -1.23456e4,
    -1.23456e3,
    -1.23456e2,
    -1.23456e1,
    -1.23456e0,
    -1.23456e-1,
    -1.23456e-2,
    -1.23456e-3,
    -1.23456e-4,
    -1.23456e-5,
    -1.23456e-6,
    -1.23456e-7,
    -1.23456e-8,
)

if full_tests:
    for type in ("e", "E", "g", "G", "n"):
        for width in ("", "4", "6", "8", "10"):
            for alignment in ("", "<", ">", "=", "^"):
                for fill in ("", "@", "0", " "):
                    for sign in ("", "+", "-", " "):
                        for prec in ("", "1", "3", "6"):
                            for num in eg_nums:
                                test_fmt("", fill, alignment, sign, "", width, prec, type, num)

# Note: We use 1.23459 rather than 1.2345 because '{:3f}'.format(1.2345)
#       rounds differently than print("%.3f", 1.2345);

f_nums = (
    0.0,
    -0.0,
    0.0001,
    0.001,
    0.01,
    0.1,
    1.0,
    10.0,
    0.0012,
    0.0123,
    0.1234,
    1.23459,
    12.3456,
    -0.0001,
    -0.001,
    -0.01,
    -0.1,
    -1.0,
    -10.0,
    -0.0012,
    -0.0123,
    -0.1234,
    -1.23459,
    -12.3456,
)

if full_tests:
    for type in ("f", "F"):
        for width in ("", "4", "6", "8", "10"):
            for alignment in ("", "<", ">", "=", "^"):
                for fill in ("", " ", "0", "@"):
                    for sign in ("", "+", "-", " "):
                        # An empty precision defaults to 6, but when uPy is
                        # configured to use a float, we can only use a
                        # precision of 6 with numbers less than 10 and still
                        # get results that compare to CPython (which uses
                        # long doubles).
                        for prec in ("1", "2", "3"):
                            for num in f_nums:
                                test_fmt("", fill, alignment, sign, "", width, prec, type, num)
                        for num in int_nums2:
                            test_fmt("", fill, alignment, sign, "", width, "", type, num)

pct_nums1 = (0.1, 0.58, 0.99, -0.1, -0.58, -0.99)
pct_nums2 = (True, False, 1, 0, -1)

if full_tests:
    type = "%"
    for width in ("", "4", "6", "8", "10"):
        for alignment in ("", "<", ">", "=", "^"):
            for fill in ("", " ", "0", "@"):
                for sign in ("", "+", "-", " "):
                    # An empty precision defaults to 6, but when uPy is
                    # configured to use a float, we can only use a
                    # precision of 6 with numbers less than 10 and still
                    # get results that compare to CPython (which uses
                    # long doubles).
                    for prec in ("1", "2", "3"):
                        for num in pct_nums1:
                            test_fmt("", fill, alignment, sign, "", width, prec, type, num)
                    for num in pct_nums2:
                        test_fmt("", fill, alignment, sign, "", width, "", type, num)
else:
    for num in pct_nums1:
        test_fmt("", "", "", "", "", "", "1", "%", num)

# We don't currently test a type of '' with floats (see the detailed comment
# in  objstr.c)



# test try handling within a viper function

# basic try-finally
@micropython.viper
def f():
    try:
        fail
    finally:
        print("finally")


try:
    f()
except NameError:
    print("NameError")

# nested try-except with try-finally
@micropython.viper
def f():
    try:
        try:
            fail
        finally:
            print("finally")
    except NameError:
        print("NameError")


f()

# check that locals written to in try blocks keep their values
@micropython.viper
def f():
    a = 100
    try:
        print(a)
        a = 200
        fail
    except NameError:
        print(a)
        a = 300
    print(a)


f()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import os, sys
from glob import glob
from re import sub
import argparse


def escape(s):
    s = s.decode()
    lookup = {
        "\0": "\\0",
        "\t": "\\t",
        "\n": '\\n"\n"',
        "\r": "\\r",
        "\\": "\\\\",
        '"': '\\"',
    }
    return '""\n"{}"'.format("".join([lookup[x] if x in lookup else x for x in s]))


def chew_filename(t):
    return {"func": "test_{}_fn".format(sub(r"/|\.|-", "_", t)), "desc": t}


def script_to_map(test_file):
    r = {"name": chew_filename(test_file)["func"]}
    with open(test_file, "rb") as test:
        script = test.readlines()

    # Test for import skip_if and inject it into the test as needed.
    if "import skip_if\n" in script:
        index = script.index("import skip_if\n")
        script.pop(index)
        script.insert(index, "class skip_if:\n")
        with open("../tests/skip_if.py") as skip_if:
            total_lines = 1
            for line in skip_if:
                stripped = line.strip()
                if not stripped or stripped.startswith(("#", '"""')):
                    continue
                script.insert(index + total_lines, "\t" + line)
                total_lines += 1
    r["script"] = escape(b"".join(script))

    with open(test_file + ".exp", "rb") as f:
        r["output"] = escape(f.read())

    return r


test_function = (
    "void {name}(void* data) {{\n"
    "  static const char pystr[] = {script};\n"
    "  static const char exp[] = {output};\n"
    '  printf("\\n");\n'
    "  upytest_set_expected_output(exp, sizeof(exp) - 1);\n"
    "  upytest_execute_test(pystr);\n"
    '  printf("result: ");\n'
    "}}"
)

testcase_struct = "struct testcase_t {name}_tests[] = {{\n{body}\n  END_OF_TESTCASES\n}};"
testcase_member = '  {{ "{desc}", {func}, TT_ENABLED_, 0, 0 }},'

testgroup_struct = "struct testgroup_t groups[] = {{\n{body}\n  END_OF_GROUPS\n}};"
testgroup_member = '  {{ "{name}", {name}_tests }},'

## XXX: may be we could have `--without <groups>` argument...
# currently these tests are selected because they pass on qemu-arm
test_dirs = (
    "basics",
    "micropython",
    "misc",
    "extmod",
    "float",
    "inlineasm",
    "qemu-arm",
)  # 'import', 'io',)
exclude_tests = (
    # pattern matching in .exp
    "basics/bytes_compare3.py",
    "extmod/ticks_diff.py",
    "extmod/time_ms_us.py",
    "extmod/uheapq_timeq.py",
    # unicode char issue
    "extmod/ujson_loads.py",
    # doesn't output to python stdout
    "extmod/ure_debug.py",
    "extmod/vfs_basic.py",
    "extmod/vfs_fat_ramdisk.py",
    "extmod/vfs_fat_fileio.py",
    "extmod/vfs_fat_fsusermount.py",
    "extmod/vfs_fat_oldproto.py",
    # rounding issues
    "float/float_divmod.py",
    # requires double precision floating point to work
    "float/float2int_doubleprec_intbig.py",
    "float/float_parse_doubleprec.py",
    # inline asm FP tests (require Cortex-M4)
    "inlineasm/asmfpaddsub.py",
    "inlineasm/asmfpcmp.py",
    "inlineasm/asmfpldrstr.py",
    "inlineasm/asmfpmuldiv.py",
    "inlineasm/asmfpsqrt.py",
    # different filename in output
    "micropython/emg_exc.py",
    "micropython/heapalloc_traceback.py",
    # don't have emergency exception buffer
    "micropython/heapalloc_exc_compressed_emg_exc.py",
    # pattern matching in .exp
    "micropython/meminfo.py",
    # needs sys stdfiles
    "misc/print_exception.py",
    # settrace .exp files are too large
    "misc/sys_settrace_loop.py",
    "misc/sys_settrace_generator.py",
    "misc/sys_settrace_features.py",
)

output = []
tests = []

argparser = argparse.ArgumentParser(
    description="Convert native MicroPython tests to tinytest/upytesthelper C code"
)
argparser.add_argument("--stdin", action="store_true", help="read list of tests from stdin")
argparser.add_argument("--exclude", action="append", help="exclude test by name")
args = argparser.parse_args()

if not args.stdin:
    if args.exclude:
        exclude_tests += tuple(args.exclude)
    for group in test_dirs:
        tests += [test for test in glob("{}/*.py".format(group)) if test not in exclude_tests]
else:
    for l in sys.stdin:
        tests.append(l.rstrip())

output.extend([test_function.format(**script_to_map(test)) for test in tests])
testcase_members = [testcase_member.format(**chew_filename(test)) for test in tests]
output.append(testcase_struct.format(name="", body="\n".join(testcase_members)))

testgroup_members = [testgroup_member.format(name=group) for group in [""]]

output.append(testgroup_struct.format(body="\n".join(testgroup_members)))

## XXX: may be we could have `--output <filename>` argument...
# Don't depend on what system locale is set, use utf8 encoding.
sys.stdout.buffer.write("\n\n".join(output).encode("utf8"))



"""Manipulate numeric data similar to numpy

`ulab` is a numpy-like module for micropython, meant to simplify and
speed up common mathematical operations on arrays. The primary goal was to
implement a small subset of numpy that might be useful in the context of a
microcontroller. This means low-level data processing of linear (array) and
two-dimensional (matrix) data.

`ulab` is adapted from micropython-ulab, and the original project's
documentation can be found at
https://micropython-ulab.readthedocs.io/en/latest/

`ulab` is modeled after numpy, and aims to be a compatible subset where
possible.  Numpy's documentation can be found at
https://docs.scipy.org/doc/numpy/index.html"""

from __future__ import annotations

from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union, overload

import ulab

_DType = int
"""`ulab.int8`, `ulab.uint8`, `ulab.int16`, `ulab.uint16`, `ulab.float` or `ulab.bool`"""

_float = float
"""Type alias of the bulitin float"""

_bool = bool
"""Type alias of the bulitin bool"""

_Index = Union[int, slice, ulab.ndarray, Tuple[Union[int, slice], ...]]

class ndarray:
    """1- and 2- dimensional ndarray"""

    def __init__(
        self,
        values: Union[ndarray, Iterable[Union[_float, _bool, Iterable[Any]]]],
        *,
        dtype: _DType = ulab.float
    ) -> None:
        """:param sequence values: Sequence giving the initial content of the ndarray.
        :param ~ulab._DType dtype: The type of ndarray values, `ulab.int8`, `ulab.uint8`, `ulab.int16`, `ulab.uint16`, `ulab.float` or `ulab.bool`

        The ``values`` sequence can either be another ~ulab.ndarray, sequence of numbers
        (in which case a 1-dimensional ndarray is created), or a sequence where each
        subsequence has the same length (in which case a 2-dimensional ndarray is
        created).

        Passing a `ulab.ndarray` and a different dtype can be used to convert an ndarray
        from one dtype to another.

        In many cases, it is more convenient to create an ndarray from a function
        like `zeros` or `linspace`.

        `ulab.ndarray` implements the buffer protocol, so it can be used in many
        places an `array.array` can be used."""
        ...
    shape: Tuple[int, ...]
    """The size of the array, a tuple of length 1 or 2"""

    size: int
    """The number of elements in the array"""

    itemsize: int
    """The size of a single item in the array"""

    strides: Tuple[int, ...]
    """Tuple of bytes to step in each dimension, a tuple of length 1 or 2"""
    def copy(self) -> ulab.ndarray:
        """Return a copy of the array"""
        ...
    def dtype(self) -> _DType:
        """Returns the dtype of the array"""
        ...
    def flatten(self, *, order: str = "C") -> ulab.ndarray:
        """:param order: Whether to flatten by rows ('C') or columns ('F')

        Returns a new `ulab.ndarray` object which is always 1 dimensional.
        If order is 'C' (the default", then the data is ordered in rows;
        If it is 'F', then the data is ordered in columns.  "C" and "F" refer
        to the typical storage organization of the C and Fortran languages."""
        ...
    def reshape(self, shape: Tuple[int, ...]) -> ulab.ndarray:
        """Returns an ndarray containing the same data with a new shape."""
        ...
    def sort(self, *, axis: Optional[int] = 1) -> None:
        """:param axis: Whether to sort elements within rows (0), columns (1), or elements (None)"""
        ...
    def tobytes(self) -> bytearray:
        """Return the raw data bytes in the ndarray"""
        ...
    def transpose(self) -> ulab.ndarray:
        """Swap the rows and columns of a 2-dimensional ndarray"""
        ...
    def __add__(self, other: Union[ndarray, _float]) -> ulab.ndarray:
        """Adds corresponding elements of the two ndarrays, or adds a number to all
        elements of the ndarray.  If both arguments are ndarrays, their sizes must match."""
        ...
    def __radd__(self, other: _float) -> ulab.ndarray: ...
    def __sub__(self, other: Union[ndarray, _float]) -> ulab.ndarray:
        """Subtracts corresponding elements of the two ndarrays, or subtracts a number from all
        elements of the ndarray.  If both arguments are ndarrays, their sizes must match."""
        ...
    def __rsub__(self, other: _float) -> ulab.ndarray: ...
    def __mul__(self, other: Union[ndarray, _float]) -> ulab.ndarray:
        """Multiplies corresponding elements of the two ndarrays, or multiplies
        all elements of the ndarray by a number.  If both arguments are ndarrays,
        their sizes must match."""
        ...
    def __rmul__(self, other: _float) -> ulab.ndarray: ...
    def __div__(self, other: Union[ndarray, _float]) -> ulab.ndarray:
        """Multiplies corresponding elements of the two ndarrays, or divides
        all elements of the ndarray by a number.  If both arguments are ndarrays,
        their sizes must match."""
        ...
    def __rdiv__(self, other: _float) -> ulab.ndarray: ...
    def __pow__(self, other: Union[ndarray, _float]) -> ulab.ndarray:
        """Computes the power (x**y) of corresponding elements of the the two ndarrays,
        or one number and one ndarray.  If both arguments are ndarrays, their sizes
        must match."""
        ...
    def __rpow__(self, other: _float) -> ulab.ndarray: ...
    def __inv__(self) -> ulab.ndarray: ...
    def __neg__(self) -> ulab.ndarray: ...
    def __pos__(self) -> ulab.ndarray: ...
    def __abs__(self) -> ulab.ndarray: ...
    def __len__(self) -> int: ...
    def __lt__(self, other: Union[ndarray, _float]) -> ulab.ndarray: ...
    def __le__(self, other: Union[ndarray, _float]) -> ulab.ndarray: ...
    def __gt__(self, other: Union[ndarray, _float]) -> ulab.ndarray: ...
    def __ge__(self, other: Union[ndarray, _float]) -> ulab.ndarray: ...
    def __iter__(self) -> Union[Iterator[ndarray], Iterator[_float]]: ...
    def __getitem__(self, index: _Index) -> Union[ndarray, _float]:
        """Retrieve an element of the ndarray."""
        ...
    def __setitem__(self, index: _Index, value: Union[ndarray, _float]) -> None:
        """Set an element of the ndarray."""
        ...

_ArrayLike = Union[ndarray, List[_float], Tuple[_float], range]
"""`ulab.ndarray`, ``List[float]``, ``Tuple[float]`` or `range`"""

int8: _DType
"""Type code for signed integers in the range -128 .. 127 inclusive, like the 'b' typecode of `array.array`"""

int16: _DType
"""Type code for signed integers in the range -32768 .. 32767 inclusive, like the 'h' typecode of `array.array`"""

float: _DType
"""Type code for floating point values, like the 'f' typecode of `array.array`"""

uint8: _DType
"""Type code for unsigned integers in the range 0 .. 255 inclusive, like the 'H' typecode of `array.array`"""

uint16: _DType
"""Type code for unsigned integers in the range 0 .. 65535 inclusive, like the 'h' typecode of `array.array`"""

bool: _DType
"""Type code for boolean values"""

def get_printoptions() -> Dict[str, int]:
    """Get printing options"""
    ...

def set_printoptions(
    threshold: Optional[int] = None, edgeitems: Optional[int] = None
) -> None:
    """Set printing options"""
    ...

def ndinfo(array: ulab.ndarray) -> None: ...
def array(
    values: Union[ndarray, Iterable[Union[_float, _bool, Iterable[Any]]]],
    *,
    dtype: _DType = ulab.float
) -> ulab.ndarray:
    """alternate constructor function for `ulab.ndarray`. Mirrors numpy.array"""
    ...

@overload
def arange(
    stop: _float, step: _float = 1, *, dtype: _DType = ulab.float
) -> ulab.ndarray: ...
@overload
def arange(
    start: _float, stop: _float, step: _float = 1, *, dtype: _DType = ulab.float
) -> ulab.ndarray:
    """
    .. param: start
      First value in the array, optional, defaults to 0
    .. param: stop
      Final value in the array
    .. param: step
      Difference between consecutive elements, optional, defaults to 1.0
    .. param: dtype
      Type of values in the array

    Return a new 1-D array with elements ranging from ``start`` to ``stop``, with step size ``step``."""
    ...

def concatenate(arrays: Tuple[ulab.ndarray], *, axis: int = 0) -> ulab.ndarray:
    """
    .. param: arrays
      tuple of ndarrays
    .. param: axis
      axis along which the arrays will be joined

    Join a sequence of arrays along an existing axis."""
    ...

def diag(a: ulab.ndarray, *, k: int = 0) -> ulab.ndarray:
    """
    .. param: a
      an ndarray
    .. param: k
      Offset of the diagonal from the main diagonal. Can be positive or negative.

    Return specified diagonals."""
    ...

def eye(
    size: int, *, M: Optional[int] = None, k: int = 0, dtype: _DType = ulab.float
) -> ulab.ndarray:
    """Return a new square array of size, with the diagonal elements set to 1
    and the other elements set to 0."""
    ...

def full(
    shape: Union[int, Tuple[int, ...]],
    fill_value: Union[_float, _bool],
    *,
    dtype: _DType = ulab.float
) -> ulab.ndarray:
    """
    .. param: shape
       Shape of the array, either an integer (for a 1-D array) or a tuple of integers (for tensors of higher rank)
    .. param: fill_value
       scalar, the value with which the array is filled
    .. param: dtype
       Type of values in the array

    Return a new array of the given shape with all elements set to 0."""
    ...

def linspace(
    start: _float,
    stop: _float,
    *,
    dtype: _DType = ulab.float,
    num: int = 50,
    endpoint: _bool = True,
    retstep: _bool = False
) -> ulab.ndarray:
    """
    .. param: start
      First value in the array
    .. param: stop
      Final value in the array
    .. param int: num
      Count of values in the array.
    .. param: dtype
      Type of values in the array
    .. param bool: endpoint
      Whether the ``stop`` value is included.  Note that even when
      endpoint=True, the exact ``stop`` value may not be included due to the
      inaccuracy of floating point arithmetic.
      If True, return (`samples`, `step`), where `step` is the spacing between samples.

    Return a new 1-D array with ``num`` elements ranging from ``start`` to ``stop`` linearly."""
    ...

def logspace(
    start: _float,
    stop: _float,
    *,
    dtype: _DType = ulab.float,
    num: int = 50,
    endpoint: _bool = True,
    base: _float = 10.0
) -> ulab.ndarray:
    """
    .. param: start
      First value in the array
    .. param: stop
      Final value in the array
    .. param int: num
      Count of values in the array. Defaults to 50.
    .. param: base
      The base of the log space. The step size between the elements in
      ``ln(samples) / ln(base)`` (or ``log_base(samples)``) is uniform. Defaults to 10.0.
    .. param: dtype
      Type of values in the array
    .. param bool: endpoint
      Whether the ``stop`` value is included.  Note that even when
      endpoint=True, the exact ``stop`` value may not be included due to the
      inaccuracy of floating point arithmetic. Defaults to True.

    Return a new 1-D array with ``num`` evenly spaced elements on a log scale.
    The sequence starts at ``base ** start``, and ends with ``base ** stop``."""
    ...

def ones(
    shape: Union[int, Tuple[int, ...]], *, dtype: _DType = ulab.float
) -> ulab.ndarray:
    """
    .. param: shape
       Shape of the array, either an integer (for a 1-D array) or a tuple of 2 integers (for a 2-D array)
    .. param: dtype
       Type of values in the array

    Return a new array of the given shape with all elements set to 1."""
    ...

def zeros(
    shape: Union[int, Tuple[int, ...]], *, dtype: _DType = ulab.float
) -> ulab.ndarray:
    """
    .. param: shape
       Shape of the array, either an integer (for a 1-D array) or a tuple of 2 integers (for a 2-D array)
    .. param: dtype
       Type of values in the array

    Return a new array of the given shape with all elements set to 0."""
    ...



# Test that calling clazz.__call__() with up to at least 3 arguments
# doesn't require heap allocation.
import micropython


class Foo0:
    def __call__(self):
        print("__call__")


class Foo1:
    def __call__(self, a):
        print("__call__", a)


class Foo2:
    def __call__(self, a, b):
        print("__call__", a, b)


class Foo3:
    def __call__(self, a, b, c):
        print("__call__", a, b, c)


f0 = Foo0()
f1 = Foo1()
f2 = Foo2()
f3 = Foo3()

micropython.heap_lock()
f0()
f1(1)
f2(1, 2)
f3(1, 2, 3)
micropython.heap_unlock()



# Array operation
# Type: bytearray, inplace operation using for. What's good about this
# method is that it doesn't require any extra memory allocation.
import bench


def test(num):
    for i in iter(range(num // 10000)):
        arr = bytearray(b"\0" * 1000)
        for i in range(len(arr)):
            arr[i] += 1


bench.run(test)



# test for+range, mostly to check optimisation of this pair

# apply args using *
for x in range(*(1, 3)):
    print(x)
for x in range(1, *(6, 2)):
    print(x)

# zero step
try:
    for x in range(1, 2, 0):
        pass
except ValueError:
    print('ValueError')

# apply args using **
try:
    for x in range(**{'end':1}):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, **{'end':1}):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, 1, **{'step':1}):
        print(x)
except TypeError:
    print('TypeError')

# keyword args
try:
    for x in range(end=1):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, end=1):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(start=0, end=1):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, 1, step=1):
        print(x)
except TypeError:
    print('TypeError')

# argument is a comprehension
try:
    for x in range(0 for i in []):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, (0 for i in [])):
        print(x)
except TypeError:
    print('TypeError')
try:
    for x in range(0, 1, (0 for i in [])):
        print(x)
except TypeError:
    print('TypeError')



import wifi
import socketpool
import struct
import time

# connect to wifi
print("Connecting to Wifi")
wifi.radio.connect("mySSID", "myPASS")
pool = socketpool.SocketPool(wifi.radio)

# make socket
print("Creating socket")
sock = pool.socket(pool.AF_INET, pool.SOCK_DGRAM)

# Fill packet
packet = bytearray(48)
packet[0] = 0b00100011  # Not leap second, NTP version 4, Client mode
NTP_TO_UNIX_EPOCH = 2208988800  # 1970-01-01 00:00:00

print("Sending packet")
sock.sendto(packet, ("pool.ntp.org", 123))

size, address = sock.recvfrom_into(packet)
print("Received packet")

seconds = struct.unpack_from("!I", packet, offset=len(packet) - 8)[0]
print("Address:", address)
print("Time:", time.localtime(seconds - NTP_TO_UNIX_EPOCH))



# test containment operator on subclass of a native type

class mylist(list):
    pass

class mydict(dict):
    pass

class mybytes(bytes):
    pass

l = mylist([1, 2, 3])
print(0 in l)
print(1 in l)

d = mydict({1:1, 2:2})
print(0 in l)
print(1 in l)

b = mybytes(b'1234')
print(0 in b)
print(b'1' in b)



# test utime resolutions

try:
    import utime
except ImportError:
    print("SKIP")
    raise SystemExit


def gmtime_time():
    return utime.gmtime(utime.time())


def localtime_time():
    return utime.localtime(utime.time())


def test():
    TEST_TIME = 2500
    EXPECTED_MAP = (
        # (function name, min. number of results in 2.5 sec)
        ("time", 3),
        ("gmtime", 3),
        ("localtime", 3),
        ("gmtime_time", 3),
        ("localtime_time", 3),
        ("ticks_ms", 15),
        ("ticks_us", 15),
        ("ticks_ns", 15),
        ("ticks_cpu", 15),
    )

    # call time functions
    results_map = {}
    end_time = utime.ticks_ms() + TEST_TIME
    while utime.ticks_diff(end_time, utime.ticks_ms()) > 0:
        utime.sleep_ms(100)
        for func_name, _ in EXPECTED_MAP:
            try:
                time_func = getattr(utime, func_name, None) or globals()[func_name]
                now = time_func()  # may raise AttributeError
            except (KeyError, AttributeError):
                continue
            try:
                results_map[func_name].add(now)
            except KeyError:
                results_map[func_name] = {now}

    # check results
    for func_name, min_len in EXPECTED_MAP:
        print("Testing %s" % func_name)
        results = results_map.get(func_name)
        if results is None:
            pass
        elif func_name == "ticks_cpu" and results == {0}:
            # ticks_cpu() returns 0 on some ports (e.g. unix)
            pass
        elif len(results) < min_len:
            print(
                "%s() returns %s result%s in %s ms, expecting >= %s"
                % (func_name, len(results), "s"[: len(results) != 1], TEST_TIME, min_len)
            )


test()



# test super with multiple inheritance

class A:
    def foo(self):
        print('A.foo')

class B:
    def foo(self):
        print('B.foo')

class C(A, B):
    def foo(self):
        print('C.foo')
        super().foo()

C().foo()



print("hello world".rfind("ll"))
print("hello world".rfind("ll", None))
print("hello world".rfind("ll", 1))
print("hello world".rfind("ll", 1, None))
print("hello world".rfind("ll", None, None))
print("hello world".rfind("ll", 1, -1))
print("hello world".rfind("ll", 1, 1))
print("hello world".rfind("ll", 1, 2))
print("hello world".rfind("ll", 1, 3))
print("hello world".rfind("ll", 1, 4))
print("hello world".rfind("ll", 1, 5))
print("hello world".rfind("ll", -100))
print("0000".rfind('0'))
print("0000".rfind('0', 0))
print("0000".rfind('0', 1))
print("0000".rfind('0', 2))
print("0000".rfind('0', 3))
print("0000".rfind('0', 4))
print("0000".rfind('0', 5))
print("0000".rfind('-1', 3))
print("0000".rfind('1', 3))
print("0000".rfind('1', 4))
print("0000".rfind('1', 5))
print("aaaaaaaaaaa".rfind("bbb", 9, 2))



import alarm
import board
import time
import microcontroller
import sdioio
import storage
import neopixel

np = neopixel.NeoPixel(board.NEOPIXEL, 1)
np[0] = (50, 0, 0)

# SD Card
sd = sdioio.SDCard(
    clock=board.SDIO_CLOCK, command=board.SDIO_COMMAND, data=board.SDIO_DATA, frequency=25000000
)
vfs = storage.VfsFat(sd)
storage.mount(vfs, "/sd")

with open("/sd/log.txt", "a") as sdc:
    temperature = microcontroller.cpu.temperature
    print("Temperature:", temperature)
    sdc.write("{}\n".format(temperature))

## USB enumeration may take 4-5s per restart
time_alarm = alarm.time.TimeAlarm(monotonic_time=time.monotonic() + 10)

np[0] = (0, 0, 0)

alarm.exit_and_deep_sleep_until_alarms(time_alarm)



# test syntax and type errors specific to viper code generation


def test(code):
    try:
        exec(code)
    except (SyntaxError, ViperTypeError, NotImplementedError) as e:
        print(repr(e))


# viper: annotations must be identifiers
test("@micropython.viper\ndef f(a:1): pass")
test("@micropython.viper\ndef f() -> 1: pass")

# unknown type
test("@micropython.viper\ndef f(x:unknown_type): pass")

# local used before type known
test(
    """
@micropython.viper
def f():
    print(x)
    x = 1
"""
)

# type mismatch storing to local
test(
    """
@micropython.viper
def f():
    x = 1
    y = []
    x = y
"""
)

# can't implicitly convert type to bool
test(
    """
@micropython.viper
def f():
    x = ptr(0)
    if x:
        pass
"""
)

# incorrect return type
test("@micropython.viper\ndef f() -> int: return []")

# can't do binary op between incompatible types
test("@micropython.viper\ndef f(): 1 + []")
test("@micropython.viper\ndef f(x:int, y:uint): x < y")

# can't load
test("@micropython.viper\ndef f(): 1[0]")
test("@micropython.viper\ndef f(): 1[x]")

# can't store
test("@micropython.viper\ndef f(): 1[0] = 1")
test("@micropython.viper\ndef f(): 1[x] = 1")
test("@micropython.viper\ndef f(x:int): x[0] = x")
test("@micropython.viper\ndef f(x:ptr32): x[0] = None")
test("@micropython.viper\ndef f(x:ptr32): x[x] = None")

# must raise an object
test("@micropython.viper\ndef f(): raise 1")

# unary ops not implemented
test("@micropython.viper\ndef f(x:int): +x")
test("@micropython.viper\ndef f(x:int): -x")
test("@micropython.viper\ndef f(x:int): ~x")

# binary op not implemented
test("@micropython.viper\ndef f(x:uint, y:uint): res = x // y")
test("@micropython.viper\ndef f(x:uint, y:uint): res = x % y")
test("@micropython.viper\ndef f(x:int): res = x in x")

# yield (from) not implemented
test("@micropython.viper\ndef f(): yield")
test("@micropython.viper\ndef f(): yield from f")

# passing a ptr to a Python function not implemented
test("@micropython.viper\ndef f(): print(ptr(1))")

# cast of a casting identifier not implemented
test("@micropython.viper\ndef f(): int(int)")



# test [...] of bytes

print(b'123'[0])
print(b'123'[1])
print(b'123'[-1])

try:
    b'123'[1] = 4
except TypeError:
    print('TypeError')

try:
    del b'123'[1]
except TypeError:
    print('TypeError')



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Copyright (c) 2017 Scott Shawcroft for Adafruit Industries
# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import os
import re
import sys

from elftools.elf.elffile import ELFFile

print()

internal_memory = [
    # Name, Start, Length
    ("RTC Fast Memory", (0x3FF9_E000, 0x4007_0000), 8 * 1024),
    ("RTC Slow Memory", (0x5000_0000,), 8 * 1024),
    ("Internal SRAM 0", (0x3FFB_0000, 0x4002_0000), 32 * 1024),
    ("Internal SRAM 1", (0x3FFB_8000, 0x4002_8000), 288 * 1024),
]


def partition_size(arg):
    if "4MB" in arg:
        return 1408 * 1024
    else:
        return 2048 * 1024


def align(n, m):
    return m * ((n + m - 1) // m)


regions = dict((name, 0) for name, _, _ in internal_memory)

# This file is the elf
with open(sys.argv[1], "rb") as stream:
    elffile = ELFFile(stream)
    for section in elffile.iter_sections():
        start = section["sh_addr"]
        size = section["sh_size"]
        offset = section["sh_offset"]
        if not size or not start:
            continue
        for name, starts, length in internal_memory:
            for mem_start in starts:
                mem_end = mem_start + length
                if start >= mem_start and start < mem_end:
                    regions[name] = max(regions.get(name, 0), size)
                    # print("# putting %s in %s (start=0x%x, size=%d)" % (section.name, name, start, size))

# This file is the sdkconfig
with open(sys.argv[2], "r") as f:
    for line in f:
        line = line.strip()
        if line.startswith("CONFIG_PARTITION_TABLE_FILENAME"):
            firmware_region = int(partition_size(line.split("=")[-1]))

# This file is the bin
used_flash = os.stat(sys.argv[3]).st_size

free_flash = firmware_region - used_flash
print(
    "{:7} bytes used, {:7} bytes free in flash firmware space out of {} bytes ({}kB).".format(
        used_flash, free_flash, firmware_region, firmware_region / 1024
    )
)
for name, mem_start, length in internal_memory:
    if name in regions:
        print(
            "{:7} bytes used, {:7} bytes free in '{}' out of {} bytes ({}kB).".format(
                regions[name], length - regions[name], name, length, length / 1024
            )
        )
print()

# Check that we have free flash space. GCC doesn't fail when the text + data
# sections don't fit in FLASH. It only counts data in RAM.
if free_flash < 0:
    print("Too little flash!!!")
    print()
    sys.exit(-1)



import board
import digitalio
import storage

switch = digitalio.DigitalInOut(board.GP4)

switch.direction = digitalio.Direction.INPUT
switch.pull = digitalio.Pull.UP

# If the switch pin is connected to ground CircuitPython can write to the drive
storage.remount("/", switch.value)



# String operations which don't require allocation
import micropython

micropython.heap_lock()

# Concatenating empty string returns original string
b"" + b""
b"" + b"1"
b"2" + b""

"" + ""
"" + "1"
"2" + ""

# If no replacements done, returns original string
"foo".replace(",", "_")

micropython.heap_unlock()



import audiobusio
import board
import digitalio
import array
import time
import math

trigger = digitalio.DigitalInOut(board.D4)
trigger.switch_to_output(True)


def mean(values):
    return sum(values) / len(values)


def normalized_rms(values):
    minbuf = int(mean(values))
    samples_sum = sum(float(sample - minbuf) * (sample - minbuf) for sample in values)

    return math.sqrt(samples_sum / len(values))


# signed 16 bit
s16 = array.array("H", [0] * 10000)

pdm = audiobusio.PDMIn(clock_pin=board.D11, data_pin=board.D12, sample_rate=24000, bit_depth=16)

print("starting read")
trigger.value = False
count = pdm.record(s16, len(s16))
trigger.value = True
print("read done")
print("recorded {} samples".format(count))
for v in s16[:count]:
    print(v)


magnitude = normalized_rms(s16)
print("magnitude", magnitude)

print("count", count)

print("done")



# test concurrent mutating access to a shared set object
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import _thread

# the shared set
se = set([-1, -2, -3, -4])

# main thread function
def th(n, lo, hi):
    for repeat in range(n):
        for i in range(lo, hi):
            se.add(i)
            assert i in se

            se.remove(i)
            assert i not in se

    with lock:
        global n_finished
        n_finished += 1


lock = _thread.allocate_lock()
n_thread = 4
n_finished = 0

# spawn threads
for i in range(n_thread):
    _thread.start_new_thread(th, (50, i * 500, (i + 1) * 500))

# busy wait for threads to finish
while n_finished < n_thread:
    pass

# check set has correct contents
print(sorted(se))



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import argparse

import os
import struct
import sys

sys.path.insert(0, "bitmap_font")
sys.path.insert(0, "../../tools/bitmap_font")

from adafruit_bitmap_font import bitmap_font

parser = argparse.ArgumentParser(description="Generate USB descriptors.")
parser.add_argument("--font", type=str, help="Font path", required=True)
parser.add_argument("--extra_characters", type=str, help="Unicode string of extra characters")
parser.add_argument(
    "--sample_file",
    type=argparse.FileType("r", encoding="utf-8"),
    help="Text file that includes strings to support.",
)
parser.add_argument("--output_c_file", type=argparse.FileType("w"), required=True)

args = parser.parse_args()


class BitmapStub:
    def __init__(self, width, height, color_depth):
        self.width = width
        self.rows = [b""] * height

    def _load_row(self, y, row):
        self.rows[y] = bytes(row)


f = bitmap_font.load_font(args.font, BitmapStub)

# Load extra characters from the sample file.
sample_characters = set()
if args.sample_file:
    for line in args.sample_file:
        # Skip comments because we add additional characters in our huffman comments.
        if line.startswith("//"):
            continue
        for c in line.strip():
            sample_characters.add(c)

# Merge visible ascii, sample characters and extra characters.
visible_ascii = bytes(range(0x20, 0x7F)).decode("utf-8")
all_characters = visible_ascii
for c in sample_characters:
    if c not in all_characters:
        all_characters += c
if args.extra_characters:
    all_characters.extend(args.extra_characters)
all_characters = "".join(sorted(set(all_characters)))
filtered_characters = all_characters

# Try to pre-load all of the glyphs. Misses will still be slow later.
f.load_glyphs(set(ord(c) for c in all_characters))

missing = 0
# Get each glyph.
for c in set(all_characters):
    if ord(c) not in f._glyphs:
        missing += 1
        filtered_characters = filtered_characters.replace(c, "")
        continue
    g = f.get_glyph(ord(c))
    if g["shift"][1] != 0:
        raise RuntimeError("y shift")

if missing > 0:
    print("Font missing", missing, "characters", file=sys.stderr)

x, y, dx, dy = f.get_bounding_box()
tile_x, tile_y = x - dx, y - dy
total_bits = tile_x * len(all_characters)
total_bits += 32 - total_bits % 32
bytes_per_row = total_bits // 8
b = bytearray(bytes_per_row * tile_y)

for x, c in enumerate(filtered_characters):
    g = f.get_glyph(ord(c))
    start_bit = x * tile_x + g["bounds"][2]
    start_y = (tile_y - 2) - (g["bounds"][1] + g["bounds"][3])
    for y, row in enumerate(g["bitmap"].rows):
        for i in range(g["bounds"][0]):
            byte = i // 8
            bit = i % 8
            if row[byte] & (1 << (7 - bit)) != 0:
                overall_bit = start_bit + (start_y + y) * bytes_per_row * 8 + i
                b[overall_bit // 8] |= 1 << (7 - (overall_bit % 8))


extra_characters = ""
for c in filtered_characters:
    if c not in visible_ascii:
        extra_characters += c

c_file = args.output_c_file

c_file.write(
    """\

#include "shared-bindings/displayio/Palette.h"
#include "supervisor/shared/display.h"

"""
)

c_file.write(
    """\
_displayio_color_t terminal_colors[2] = {
    {
        .rgb888 = 0x000000,
        .rgb565 = 0x0000,
        .luma = 0x00,
        .chroma = 0
    },
    {
        .rgb888 = 0xffffff,
        .rgb565 = 0xffff,
        .luma = 0xff,
        .chroma = 0
    },
};

displayio_palette_t supervisor_terminal_color = {
    .base = {.type = &displayio_palette_type },
    .colors = terminal_colors,
    .color_count = 2,
    .needs_refresh = false
};
"""
)

c_file.write(
    """\
displayio_tilegrid_t supervisor_terminal_text_grid = {{
    .base = {{ .type = &displayio_tilegrid_type }},
    .bitmap = (displayio_bitmap_t*) &supervisor_terminal_font_bitmap,
    .pixel_shader = &supervisor_terminal_color,
    .x = 16,
    .y = 0,
    .pixel_width = {1},
    .pixel_height = {2},
    .bitmap_width_in_tiles = {0},
    .tiles_in_bitmap = {0},
    .width_in_tiles = 1,
    .height_in_tiles = 1,
    .tile_width = {1},
    .tile_height = {2},
    .tiles = NULL,
    .partial_change = false,
    .full_change = false,
    .hidden = false,
    .hidden_by_parent = false,
    .moved = false,
    .inline_tiles = false,
    .in_group = true
}};
""".format(
        len(all_characters), tile_x, tile_y
    )
)

c_file.write(
    """\
const uint32_t font_bitmap_data[{}] = {{
""".format(
        bytes_per_row * tile_y // 4
    )
)

for i, word in enumerate(struct.iter_unpack(">I", b)):
    c_file.write("0x{:08x}, ".format(word[0]))
    if (i + 1) % (bytes_per_row // 4) == 0:
        c_file.write("\n")

c_file.write(
    """\
};
"""
)

c_file.write(
    """\
displayio_bitmap_t supervisor_terminal_font_bitmap = {{
    .base = {{.type = &displayio_bitmap_type }},
    .width = {},
    .height = {},
    .data = (size_t*) font_bitmap_data,
    .stride = {},
    .bits_per_value = 1,
    .x_shift = 5,
    .x_mask = 0x1f,
    .bitmask = 0x1,
    .read_only = true
}};
""".format(
        len(all_characters) * tile_x, tile_y, bytes_per_row / 4
    )
)


c_file.write(
    """\
const fontio_builtinfont_t supervisor_terminal_font = {{
    .base = {{.type = &fontio_builtinfont_type }},
    .bitmap = &supervisor_terminal_font_bitmap,
    .width = {},
    .height = {},
    .unicode_characters = (const uint8_t*) "{}",
    .unicode_characters_len = {}
}};
""".format(
        tile_x, tile_y, extra_characters, len(extra_characters.encode("utf-8"))
    )
)

c_file.write(
    """\
terminalio_terminal_obj_t supervisor_terminal = {
    .base = { .type = &terminalio_terminal_type },
    .font = &supervisor_terminal_font,
    .cursor_x = 0,
    .cursor_y = 0,
    .tilegrid = &supervisor_terminal_text_grid
};
"""
)



# test list.__iadd__ and list.extend (they are equivalent)

l = [1, 2]
l.extend([])
print(l)

l.extend([3])
print(l)

l.extend([4, 5])
print(l)

l.extend(range(6, 10))
print(l)

l.extend("abc")
print(l)

l = [1, 2]
l += []
print(l)

l += [3]
print(l)

l += [4, 5]
print(l)

l += range(6, 10)
print(l)

l += "abc"
print(l)



#!/usr/bin/env python

# This pre-processor parses provided objects' c files for
# MP_REGISTER_MODULE(module_name, obj_module, enabled_define)
# These are used to generate a header with the required entries for
# "mp_rom_map_elem_t mp_builtin_module_table[]" in py/objmodule.c

from __future__ import print_function

import re
import io
import os
import argparse


pattern = re.compile(r"[\n;]\s*MP_REGISTER_MODULE\((.*?),\s*(.*?),\s*(.*?)\);", flags=re.DOTALL)


def find_c_file(obj_file, vpath):
    """Search vpaths for the c file that matches the provided object_file.

    :param str obj_file: object file to find the matching c file for
    :param List[str] vpath: List of base paths, similar to gcc vpath
    :return: str path to c file or None
    """
    c_file = None
    relative_c_file = os.path.splitext(obj_file)[0] + ".c"
    relative_c_file = relative_c_file.lstrip("/\\")
    for p in vpath:
        possible_c_file = os.path.join(p, relative_c_file)
        if os.path.exists(possible_c_file):
            c_file = possible_c_file
            break

    return c_file


def find_module_registrations(c_file):
    """Find any MP_REGISTER_MODULE definitions in the provided c file.

    :param str c_file: path to c file to check
    :return: List[(module_name, obj_module, enabled_define)]
    """
    global pattern

    if c_file is None:
        # No c file to match the object file, skip
        return set()

    with io.open(c_file, encoding="utf-8") as c_file_obj:
        return set(re.findall(pattern, c_file_obj.read()))


def generate_module_table_header(modules):
    """Generate header with module table entries for builtin modules.

    :param List[(module_name, obj_module, enabled_define)] modules: module defs
    :return: None
    """

    # Print header file for all external modules.
    mod_defs = []
    print("// Automatically generated by makemoduledefs.py.\n")
    print('#include "py/mpconfig.h"')
    for module_name, obj_module, enabled_define in modules:
        mod_def = "MODULE_DEF_{}".format(module_name.upper())
        mod_defs.append(mod_def)
        print(
            (
                "#if ({enabled_define})\n"
                "    extern const struct _mp_obj_module_t {obj_module};\n"
                "    #define {mod_def} {{ MP_ROM_QSTR({module_name}), MP_ROM_PTR(&{obj_module}) }},\n"
                "#else\n"
                "    #define {mod_def}\n"
                "#endif\n"
            ).format(
                module_name=module_name,
                obj_module=obj_module,
                enabled_define=enabled_define,
                mod_def=mod_def,
            )
        )

    print("\n#define MICROPY_REGISTERED_MODULES \\")

    for mod_def in mod_defs:
        print("    {mod_def} \\".format(mod_def=mod_def))

    print("// MICROPY_REGISTERED_MODULES")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--vpath", default=".", help="comma separated list of folders to search for c files in"
    )
    parser.add_argument("files", nargs="*", help="list of c files to search")
    args = parser.parse_args()

    vpath = [p.strip() for p in args.vpath.split(",")]

    modules = set()
    for obj_file in args.files:
        c_file = find_c_file(obj_file, vpath)
        modules |= find_module_registrations(c_file)

    generate_module_table_header(sorted(modules))


if __name__ == "__main__":
    main()



# Function call overhead test
# Perform the same trivial operation as global function call
import bench


def f(x):
    return x + 1


def test(num):
    for i in iter(range(num)):
        a = f(i)


bench.run(test)
